import datetime


def print_head():
    program_name = "automated Reaction Search Tool (aRST) V2.0"
    timestamp = datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S")
    tot_width = 80
    border_width = 60
    side_padding = (tot_width - border_width) // 2
    side = " " * side_padding
    border = "=" * border_width

    print(side + border)
    print(side + f"|{' ' * (border_width - 2)}|")
    print(side + f"|{' ' * (border_width - 2)}|")
    print(side + f"|{program_name.center(border_width - 2)}|")
    print(side + f"|{' ' * (border_width - 2)}|")
    print(side + f"|{' ' * (border_width - 2)}|")
    print(side + border)

    print(side+"# Author:             Y. Chen")
    print(side+"# Institute:          Institute of Physical Chemistry")
    print(side+"#                     RWTH Aachen University")
    print(side+"# Version date:       03.2025")
    print(side+f"# Job running at:     {timestamp}")
    print(side + border)
    print()


def print_args(args):
    string = " "
    if args.explore:
        string = "Job type: reaction exploartion"
    elif args.afo_features:
        string = "Job type: reactivity assesment"
    elif args.align:
        string = "Job type: reactants alignment"
    print(string.center(80))


def print_parser(data):
    print("*read input:")
    for section, subkeys in data.items():
        # Start with section header
        line = f"[{section}]:"
        indent = " " * 4  # 4-space indent for continuation lines
        first_item_in_line = True  # Track if we're at the start of a new line

        # Add key-value pairs one by one, checking line length
        for k, v in subkeys.items():
            # Add comma only if it's not the first item in the line
            item = f", {k}={v}" if not first_item_in_line else f" {k}={v}"

            # If adding this item would exceed 60 chars, start a new line
            if len(line) + len(item) > 60:
                print(line)
                line = indent + f"{k}={v}"  # Start new line with indent (no comma)
                first_item_in_line = True  # Next item is the first in this line
            else:
                line += item
                first_item_in_line = False  # Next item is not the first

        print(line)  # Print the final line for the section
    print()
    print()


def print_simulation(reactants, at_l, charge_l, nel_l, reactionpath, count, totnel, totcharge, u, nat_1=None):
    totw = 80
    border_width = 60
    col1 = 15
    col2 = 15
    col3 = 15
    col4 = 15
    side_padding = (totw - border_width) // 2
    side = " " * side_padding

    print(side + "=" * border_width)
    title = f"Path Guess Scan {reactionpath}_{count}"
    print(title.center(totw))
    print(side + "-" * border_width)
    print(side + f"{'struc'.center(col1)}{'charge'.center(col2)}{'nel'.center(col3)}{'react atom'.center(col4)}")
    print(side + "-" * border_width)

    if len(at_l) > 1:
        for reactant, at1, at2, charge, nel in zip(reactants, at_l[0], at_l[1], charge_l, nel_l):
            atoms = f"({at1},{at2})"
            print(side + f"{str(reactant).center(col1)}"
                         f"{str(charge).center(col2)}"
                         f"{str(nel).center(col3)}"
                         f"{str(atoms).center(col4)}")

    else:
        for reactant, at, charge, nel in zip(reactants, at_l[0], charge_l, nel_l):
            print(side + f"{str(reactant).center(col1)}"
                         f"{str(charge).center(col2)}"
                         f"{str(nel).center(col3)}"
                         f"{str(at).center(col4)}")

    print(side + "-" * border_width)
    at1 = int(at_l[0][0])
    at2 = int(at_l[0][1])
    if nat_1:
        at2 += nat_1
    if len(at_l) > 1:
        at3 = int(at_l[1][0])
        at4 = int(at_l[1][1])
        if nat_1:
            at4 += nat_1
        string = f"({at1, at2},{at3, at4}"
    else:
        string = f"{at1, at2}"
    print(side + f"{f'P(uhf={u})'.center(col1)}"
                 f"{str(totcharge).center(col2)}"
                 f"{str(totnel).center(col3)}"
                 f"{str(string).center(col4)}")
    print(side + "=" * border_width)
    print()


def print_simulation_details(status,key=None, atl=None, sim_uhf=None, otherkey=None, E_sp=None):
    totw = 80
    border_width = 60
    col1 = 10
    col2 = 15
    col3 = 5
    col4 = 30
    side_padding = (totw - border_width) // 2
    side = " " * side_padding
    comments = None

    if status == 0:  # print head
        print(side + "=" * border_width)
        title = f"Reaction Simulation Summary"
        print(title.center(totw))
        print(side + "-" * border_width)
        print(side + f"{'scan':^{col1}} {'atl':^{col2}} {'uhf_sim':^{col3}} {'status':^{col4}}")
        print(side + "-" * border_width)
        return
    elif status == 5:  # print tail
        print(side + "=" * border_width)
        return
    elif status == 1:
        # reactants = products
        comments = "R=P"
    elif status == 2:
        # rxn exist
        comments = f"rxn exist in {otherkey}"

    elif status == 3:
        # rxn exist but different unreact
        comments = f" semi-new rxn: {otherkey}"
    elif status == 4:
        # new reaction
        comments = f" new rxn: {otherkey}, ΔE_sp ={E_sp:.2f}"
    elif status==6:
        # unknow in rmb or pmb
        comments = f" UMBnew rxn: {otherkey}, ΔE_sp ={E_sp:.2f}"
    print(side + f"{str(key).center(col1)}"
                 f"{str(atl).center(col2)}"
                 f"{str(sim_uhf).center(col3)}"
                 f"{str(comments).center(col4)}")


def print_jobhead(jobhead):
    string = "****"
    stringout = string + " " + jobhead + " " + string
    print(stringout.center(80))


def print_react_l(reactants,react_l,rtype="interA"):
    totw = 80
    border_width = 60
    col1 = 15
    col2 = 15
    col3 = 35
    side_padding = (totw - border_width) // 2
    side = " " * side_padding
    print(side + "*" * border_width)
    title = f"{rtype} Reactivity Assessment"
    print(title.center(totw))
    print()
    print(side + f"Reactants: {reactants}")
    print(side + f"tot.atom pairs: {len(react_l)}")
    print(side + "-" * border_width)
    print(side + f"{'struc pair':^{col1}} {'atom pair':^{col2}} {'AFO gap':^{col3}}")
    print(side + "-" * border_width)
    for entry in react_l[:20]:
        struc_l = entry["struc_combi"] \
            if "struc_combi" in entry.keys() else entry["struc"]
        at_l = entry["at_combi"]
        gap = entry["gap"]
        strgap = f"{gap:.2f}" if isinstance(gap, float) else "np.inf"
        print(side + f"{str(struc_l).center(col1)}"
                     f"{str(at_l).center(col2)}"
                     f"{strgap.center(col3)}")
    print(side + "*" * border_width)



import sys
import toml
import threading
import os
from os.path import join, exists
from aRST.script.toolbox import create_folder, mydict, read_coord, get_strucinfo_from_path, write_xyz, get_all_keys
from aRST.script.callsys import CallxTB, CallORCA
from aRST.script.check_input import checking_parameter, checking_env


class Setting:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, infile=None):
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._instance._load_config(infile)
        return cls._instance

    def _load_config(self, infile):
        with open(infile, "r") as f:
            self.config = toml.load(f)
        self.config = mydict(self.config)
        # changing all str parameter to lower cases
        all_keys = get_all_keys(self.config)
        for key in all_keys:
            val = self.get_value(key)
            if isinstance(val, str) and not "geom" in key:
                self.set_value(key, val.lower())

        # checking up parameter format

        self.config = checking_parameter(self.config)
        checking_env(self.config)

        self.wd0 = self.get_value("searching_setting.general.wd", os.path.abspath(os.path.dirname(infile)))
        self.scanwd = join(self.wd0, "scan")
        self.strucwd = join(self.wd0, "struc")
        self.bufferwd = join(self.wd0, "buffer_struc")
        # clean up buffer struc
        if exists(self.bufferwd):
            import shutil
            shutil.rmtree(self.bufferwd)

    def get_value(self, target_key, default_value=None):
        value = mydict(self.config).get_value(target_key, default_value)
        return value

    def set_value(self, target_key, new_value):
        mydict(self.config).set_value(target_key, new_value)


class Allstruc:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls, Setting=None):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
                    cls._instance._initialize(Setting)
        return cls._instance

    def _initialize(self, Setting):
        self.number_of_struc = mydict(
            {})  # only used in nanoreactor for recording change of number of structures in sys
        self.number_of_struc["0"] = []
        # self.number_of_struc = {"layer":[{"strucid": strucid,
        #                                   "num": num}, ...]}
        self.data = mydict({})
        self.reactants_ini = []
        # read geom input
        ## reactants
        reactants = [_ for _ in Setting.get_value("geom").keys() if "reactant" in _]
        for id, struc in enumerate(reactants):
            coord = read_coord(join(Setting.wd0, Setting.get_value(f"geom.{struc}.name")))
            c = Setting.get_value(f"geom.{struc}.charge")
            m = Setting.get_value(f"geom.{struc}.multiplicity")
            if all(v is not None for v in [coord, c, m]):
                # set up new id anyhow
                self.data.set_value(f"{str(id)}.struc_info.general.coord", coord)
                self.data.set_value(f"{str(id)}.struc_info.general.charge", c)
                self.data.set_value(f"{str(id)}.struc_info.general.multiplicity", m)
            else:
                raise ValueError("Can't read reactants info from input, check if correct coord name, charge, m!")
            nstruc = Setting.get_value(f"geom.{struc}.number", 1)
            self.data.set_value(f"{str(id)}.struc_info.general.number", nstruc)
            self.number_of_struc["0"].append({"strucid": str(id), "num": int(nstruc)})
            for _ in range(nstruc):
                self.reactants_ini.append(str(id))

        ## ref products
        self.refproducts = []
        self.data_refp = mydict({})
        products = [_ for _ in Setting.get_value("geom").keys() if "product" in _]
        for id, struc in enumerate(products):
            coord = read_coord(join(Setting.wd0, Setting.get_value(f"geom.{struc}.name")))
            c = Setting.get_value(f"geom.{struc}.charge")
            m = Setting.get_value(f"geom.{struc}.multiplicity")
            if all(v is not None for v in [coord, c, m]):
                self.data_refp.set_value(f"{str(id)}.struc_info.general.coord", coord)
                self.data_refp.set_value(f"{str(id)}.struc_info.general.charge", c)
                self.data_refp.set_value(f"{str(id)}.struc_info.general.multiplicity", m)
            else:
                raise ValueError("Can't read refproducts info from input, check if correct coord name, charge, m!")
            for _ in range(Setting.get_value(f"geom.{struc}.number", 1)):
                self.refproducts.append(str(id))

        # collect struc refine info
        if Setting.get_value(f"read.general.jumprefine", False):
            # direct read struc refine information from foders 0, 1, etc
            for id, struc in enumerate(reactants):
                try:
                    struc_info = get_strucinfo_from_path(join(Setting.wd0, "struc", str(id)))
                    self.set_value_from_struc_info(str(id), struc_info)
                except:
                    print(f"{str(id)} strucinfo doesn't exist!, start to refine it")
                    coord = read_coord(join(Setting.wd0, Setting.get_value(f"geom.{struc}.name")))
                    c = Setting.get_value(f"geom.{struc}.charge")
                    m = Setting.get_value(f"geom.{struc}.multiplicity")
                    struc_info = self.refine_struc_info(coord, c, m, Setting, str(id))
                    self.set_value_from_struc_info(str(id), struc_info)


        else:
            for id, struc in enumerate(reactants):
                coord = read_coord(join(Setting.wd0, Setting.get_value(f"geom.{struc}.name")))
                c = Setting.get_value(f"geom.{struc}.charge")
                m = Setting.get_value(f"geom.{struc}.multiplicity")
                struc_info = self.refine_struc_info(coord, c, m, Setting, str(id))
                self.set_value_from_struc_info(str(id), struc_info)

    def set_value_from_struc_info(self, strucid, struc_info):
        # test if this reactant is already in data
        mb = struc_info["mb"]
        for searchid in list(self.data.keys()):
            if searchid != strucid:
                search_mb = self.data.get_value(f"{searchid}.struc_info.general.mb")
                if mb == search_mb:
                    del self.data[strucid]
                    self.reactants_ini.remove(strucid)
                    self.reactants_ini.append(searchid)
                    return
        # check if need to change current strucid
        for _ in range(int(strucid)):
            if str(_) not in list(self.data.keys()):
                del self.data[strucid]
                strucid = str(_)
                break

        self.data.set_value(f"{strucid}.struc_info.general.coord", struc_info["coord"])
        self.data.set_value(f"{strucid}.struc_info.general.nel", struc_info["nel"])
        self.data.set_value(f"{strucid}.struc_info.general.multiplicity", struc_info["m"])
        self.data.set_value(f"{strucid}.struc_info.general.charge", struc_info["c"])
        self.data.set_value(f"{strucid}.struc_info.general.atcharge", struc_info["atcharge"])
        self.data.set_value(f"{strucid}.struc_info.general.mb", struc_info["mb"])
        self.data.set_value(f"{strucid}.struc_info.general.atomic_connection",
                            (struc_info["constrain_l"], struc_info["ele_l"]))

        self.data.set_value(f"{strucid}.struc_info.energy.sp_gfn", struc_info["Esp_gfn"])
        self.data.set_value(f"{strucid}.struc_info.energy.sp_dft", struc_info["Esp_orca"])
        # self.data.set_value(f"{strucid}.struc_info.energy.trv", struc_info["Etrv_gfn"])
        # self.data.set_value(f"{strucid}.struc_info.energy.solv", struc_info["Esolv_gfn"])

    @staticmethod
    def refine_struc_info(coord, charge, m, Setting, strucid=0, wd0=None):

        #########  OPT  #########

        # do gfn opt if required
        if Setting.get_value("read.general.gfn2opt", False) and coord.shape[0] > 1:
            if not wd0:
                wd = create_folder(join(Setting.strucwd, strucid, "xtb"))
            else:
                wd = create_folder(join(wd0, "xtb"))
            write_xyz(coord, "coord")
            staus = CallxTB(wd, custom_settings={"charge": charge, "uhf": m - 1}).opt()
            if staus == 0:
                coord = read_coord(join(wd, "xtbopt.xyz"))
            else:
                print(f'Warning: struc {strucid} opt at xtb failed, continue with unopt struc')

        # do DFT opt if required
        forcedftopt = Setting.get_value("read.general.dftopt", False)
        if (forcedftopt or Setting.get_value("orca_setting.dftopt", False)) and coord.shape[0] > 1:
            if not wd0:
                wd = create_folder(join(Setting.strucwd, strucid, "orca"))
            else:
                wd = create_folder(join(wd0, "orca"))
            write_xyz(coord, "coord")
            staus = CallORCA(wd, custom_settings={"charge": charge, "m": m, "caltyp": "SP OPT"}).call()
            if staus == 0:
                coord = read_coord(join(wd, "orca.xyz"))
            else:
                print(f'Warning: struc {strucid} opt at ORCA failed, continue with unopt struc')

        #########  SP  #########

        ## xtb sp
        if not wd0:
            if not exists(join(Setting.strucwd, strucid, "xtb", "xtbopt.xyz")):
                wd = create_folder(join(Setting.strucwd, strucid, "xtb"))
                write_xyz(coord, "coord")
                staus = CallxTB(wd, custom_settings={"charge": charge, "uhf": m - 1}).sp()
                if staus != 0:
                    print(f'Warning: struc {strucid} sp at xtb failed, continue anyhow')
        else:
            wd = create_folder(join(wd0, "xtb"))
            write_xyz(coord, "coord")
            staus = CallxTB(wd, custom_settings={"charge": charge, "uhf": m - 1}).sp()
            if staus != 0:
                print(f'Warning: struc {strucid} sp at xtb failed, continue anyhow')

        ## orca sp
        sp_orca = None
        if not (forcedftopt or Setting.get_value("orca_setting.dftopt", False)) or coord.shape[0] == 1:
            if not wd0:
                wd = create_folder(join(Setting.strucwd, strucid, "orca"))
            else:
                wd = create_folder(join(wd0, "orca"))
            write_xyz(coord, "coord")
            staus = CallORCA(wd, custom_settings={"charge": charge, "m": m}).call()
            if staus != 0:
                print(f'Warning: struc {strucid} sp at ORCA failed, continue anyhow')

        # #########  TRV  #########
        # wd = create_folder(join(Setting.strucwd, strucid, "xtb", "trv"))
        # write_xyz(coord, "coord")
        # staus = CallxTB(wd, custom_settings={"charge": charge, "uhf": m - 1}).trv()
        # if staus != 0:
        #     print(f'Warning: struc {strucid} trv failed, continue anyhow')

        #########  SOLV  #########
        # wd = create_folder(join(Setting.strucwd, strucid, "xtb", "solv"))
        # write_xyz(coord, "coord")
        # staus = CallxTB(wd, custom_settings={"charge": charge, "uhf": m - 1}).solv()
        # if staus != 0:
        #     print(f'Warning: struc {strucid} solv failed, continue anyhow')

        #########  get res  #########
        if not wd0:
            struc_info = get_strucinfo_from_path(join(Setting.strucwd, strucid))

            return struc_info

    def get_value(self, target_key, default_value=None):
        value = self.data.get_value(target_key, default_value)
        return value

    def set_value(self, target_key, new_value):
        self.data.set_value(target_key, new_value)

    def compare_value(self, key1, key2):
        return self.data.compare_value(key1, key2)


class Allrxn:
    _instance = None
    _lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
                    cls._instance.data = mydict({})
        return cls._instance

    def get_value(self, target_key, default_value=None):
        value = self.data.get_value(target_key, default_value)
        return value

    def set_value(self, target_key, new_value):
        self.data.set_value(target_key, new_value)

    def compare_value(self, key1, key2):
        return self.data.compare_value(key1, key2)

    def get_all_keys(self):
        return list(self.data.keys())

    def get_layer_keys(self, layer=0):
        # cycle0: 0_0, 0_1, 0_2....
        # cycle1: 0_0_0, 0_0_1,0_0_2......
        allkeys = self.get_all_keys()
        return [_ for _ in allkeys if len(_.split("_")) == layer + 1]


setting = None
allstruc = None
allrxn = Allrxn()


def initialize(setting_instance):
    """Initialize global setting and allstruc with the provided Setting instance."""
    global setting, allstruc
    setting = setting_instance  # Assign the passed Setting instance
    allstruc = Allstruc(setting)



import os
import numpy as np
import pandas as pd
from os.path import join, exists
import networkx as nx
from aRST.script.element import element


class mydict(dict):
    def __init__(self, data=None):
        if data is None:
            data = {}
        super().__init__(data)

    def get_value(self, target_key, default_value=None):
        keys = target_key.split('.')
        current_data = self

        for key in keys:
            if isinstance(current_data, dict) and key in current_data:
                current_data = current_data[key]
            elif isinstance(current_data, list):
                try:
                    index = int(key)
                    current_data = current_data[index]
                except (ValueError, IndexError):
                    return current_data
            else:
                return default_value

        return current_data

    def set_value(self, target_key, new_value):
        keys = target_key.split('.')
        current_data = self

        for key in keys[:-1]:
            if isinstance(current_data, dict):
                if key not in current_data:
                    current_data[key] = {}
                current_data = current_data[key]
            else:
                raise KeyError(f"Invalid key '{key}' in the dictionary.")

        last_key = keys[-1]
        current_data[last_key] = new_value

    def compare_value(self, key1, key2):
        value1 = self.get_value(key1)
        value2 = self.get_value(key2)
        if isinstance(value1, str) and isinstance(value2, str):
            return value1 == value2
        if isinstance(value1, (int, float)) and isinstance(value2, (int, float)):
            return round(value1, 2) == round(value2, 2)
        return False


def get_all_keys(d, parent_key=""):
    keys = []

    if isinstance(d, dict):
        for k, v in d.items():
            full_key = f"{parent_key}.{k}" if parent_key else k
            keys.extend(get_all_keys(v, full_key))
    elif isinstance(d, list):
        for idx, item in enumerate(d):
            full_key = f"{parent_key}.{idx}" if parent_key else str(idx)
            keys.extend(get_all_keys(item, full_key))
    else:
        # It's a leaf value (non-dict, non-list)
        keys.append(parent_key)
    return keys


def read_coord(coordpath):
    coord = pd.read_table(coordpath, header=None, skiprows=2, sep='\s+') if exists(coordpath) else None
    return coord


def read_bo(bopath):
    bo = pd.read_csv(bopath, header=None, sep='\s+', dtype={0: int, 1: int}) \
        if exists(bopath) and os.path.getsize(bopath) != 0 else None
    return bo


def read_afo_csv(afopath):
    afo = pd.read_csv(afopath, index_col=False) if exists(afopath) else None
    if afo is not None:
        afo.columns = afo.columns.str.strip()
        afo.insert(0, 'index', range(len(afo)), allow_duplicates=False)
        afo.set_index("index")
        if 'HOAO_a (eV)' in afo.columns:
            afo = afo[['index',
                       'HOAO_a (eV)', 'HOAO_b (eV)',
                       'LUAO_a (eV)', 'LUAO_b (eV)', 'Atom']]
        else:
            afo = afo[['index',
                       'HOAO (eV)', 'LUAO (eV)', 'Atom']]
            afo.rename(columns={'HOAO (eV)': 'HOAO_a (eV)'})
            afo.rename(columns={'LUAO (eV)': 'LUAO_a (eV)'})
            afo.insert(len(afo.columns), "HOAO_b (eV)", afo.loc[:, "HOAO_a (eV)"].copy())
            afo.insert(len(afo.columns), "LUAO_b (eV)", afo.loc[:, "LUAO_a (eV)"].copy())
    return afo


def read_atcharge(atchargepath):
    atcharge = pd.read_table(atchargepath, header=None, sep='\s+') if exists(atchargepath) else None
    return atcharge


def get_strucinfo_from_path(strucid_path):
    from aRST.script.callsys import CallxTB, CallORCA, CallMolBar
    # determine coord
    if exists(join(strucid_path, "orca", "orca.xyz")):
        coord = read_coord(join(strucid_path, "orca", "orca.xyz"))
    elif exists(join(strucid_path, "orca", "coord.xyz")):
        coord = read_coord(join(strucid_path, "orca", "coord.xyz"))
    elif exists(join(strucid_path, "xtb", "xtbopt.xyz")):
        coord = read_coord(join(strucid_path, "orca", "xtbopt.xyz"))

    # at least it should have xtb sp res

    # elif exists(join(strucid_path, "xtb", "coord.xyz")):
    #     coord = read_coord(join(strucid_path, "orca", "coord.xyz"))
    else:
        raise ValueError(f"{strucid_path} strucinfo doesn't exist!")

    xtb = CallxTB(join(strucid_path, "xtb"))
    charge = xtb.get_charge_from_out()
    ele_l = coord.iloc[:, 0].tolist()
    nel = int(sum([element[_] for _ in ele_l])) - charge
    atcharge = read_atcharge(join(strucid_path, "charges"))
    Esp_gfn = xtb.get_Esp_from_out()
    # Etrv_gfn = CallxTB(join(strucid_path, "xtb", "trv")).get_Etrv()
    Esolv_gfn = CallxTB(join(strucid_path, "xtb", "solv")).get_Esolv()
    constrain_l, ele_l = xtb.get_topo()

    Esp_orca = CallORCA(join(strucid_path, "orca")).get_energy()
    m = CallORCA(join(strucid_path, "orca")).get_m()

    mb = CallMolBar(coord).get_mb()
    return {"coord": coord, "nel": nel, "m": m, "c": charge, "atcharge": atcharge,
            "Esp_gfn": Esp_gfn,
            # "Etrv_gfn": Etrv_gfn,
            # "Esolv_gfn": Esolv_gfn,
            "constrain_l": constrain_l, "ele_l": ele_l, "Esp_orca": Esp_orca, "mb": mb}


def create_folder(wd):
    wd = os.path.abspath(wd)
    if not exists(wd):
        os.makedirs(wd)
        os.chdir(wd)
    elif os.path.isdir(wd):
        os.chdir(wd)
    else:
        os.remove(wd)
        os.makedirs(wd)
        os.chdir(wd)
    return wd


def get_random_folder_name():
    import random
    import string
    random_name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5))
    return random_name


def write_xyz(xyz, name):
    pwd = os.getcwd()
    # xyz.round(10)
    nat = xyz.shape[0]
    with open(f'{pwd}/{name}.xyz', 'w') as f:
        f.write(f'{nat}\n')
        f.write('\n')
        for i in range(nat):
            at, x, y, z = xyz.iloc[i, 0], xyz.iloc[i, 1], xyz.iloc[i, 2], xyz.iloc[i, 3]
            f.write('%2s %12.4f %12.4f %12.4f\n' % (at, x, y, z))
        f.write('\n')


class FragmentAnalyzer:
    def __init__(self, coord, bo, atcharge):

        # supra struc info
        self.Snat = coord.shape[0]
        self.Sbo = bo
        self.Satcharge = atcharge
        self.Scharge = np.round(atcharge.iloc[:, 0].sum())
        self.Scoord = coord
        self.Sele_l = coord.iloc[:, 0].tolist()
        self.Snel = int(sum([element[_] for _ in self.Sele_l])) - self.Scharge

        # build up graph
        self.build_cn_matrix()

        self.define_graph()
        # self.degree = [deg for node, deg in self.Graph.degree()]

        # fragment struc info
        self.fragments = [list(_) for _ in nx.connected_components(self.Graph)]

        self.fragment_data = self.get_fragment_info()

        # check if nel doesn't change
        if not self.check_nel():
            self.status = 1
        else:
            self.status = 0

    def build_cn_matrix(self):
        self.Mcn = np.zeros((self.Snat, self.Snat), dtype=int)
        # build up cn matrix base on wbo file
        if self.Sbo is not None:
            for i in range(self.Sbo.shape[0]):
                if self.Sbo.iloc[i, 2] > 0.3:
                    at1, at2 = self.Sbo.iloc[i, :2]
                    at1, at2 = int(at1), int(at2)
                    self.Mcn[at1 - 1, at2 - 1] = self.Mcn[at2 - 1, at1 - 1] = 1

    def define_graph(self):
        self.Graph = nx.Graph()
        for i in range(self.Snat):
            self.Graph.add_node(i)
            for j in range(i + 1, self.Snat):
                if self.Mcn[i, j]:
                    self.Graph.add_edge(i, j)
        G_sorted = nx.Graph()
        for node in sorted(self.Graph.nodes):
            G_sorted.add_node(node)
        for u, v, data in self.Graph.edges(data=True):
            G_sorted.add_edge(min(u, v), max(u, v), weight=data["weight"] if "weight" in data.keys() else 1)
        self.Graph = G_sorted

    def get_fragment_info(self):
        charge_l_flo = []
        fragment_data = []
        for fragment in self.fragments:
            atcharge = self.Satcharge.iloc[list(fragment)]
            charge_flo = atcharge.sum()
            charge_l_flo.append(float(charge_flo))
            fragment_data.append({
                "charge_flo": charge_flo, "atcharge": atcharge})

        # Convert float fragment charges to integers while conserving total charge
        int_fragment_charges = self.adjust_fragment_charge(charge_l_flo)

        # Update fragment data with integer values
        for i, frag in enumerate(fragment_data):
            coord = self.Scoord.iloc[self.fragments[i]]
            ele_l = coord.iloc[:, 0].tolist()
            charge_int = int(int_fragment_charges[i])
            nel = int(sum([element[_] for _ in ele_l]) - charge_int)
            frag["coord"] = coord
            frag["charge_int"] = charge_int
            frag["nel"] = nel
        return fragment_data

    def check_nel(self):
        tmpnel = sum([_["nel"] for _ in self.fragment_data])

        if tmpnel == self.Snel:
            return True
        else:
            return False

    def adjust_fragment_charge(self, charge_l_flo):

        int_charges = np.round(charge_l_flo).astype(int)  # Round to the nearest integer
        charge_difference = self.Scharge - np.sum(int_charges)  # Compute discrepancy

        if charge_difference != 0:
            residuals = np.array(charge_l_flo) - int_charges  # Find decimal parts
            adjustment_indices = np.argsort(-residuals)  # Sort by highest decimal part

            for i in range(int(abs(charge_difference))):
                idx = adjustment_indices[i]
                int_charges[idx] += np.sign(charge_difference)  # Adjust by +1 or -1
        return list(int_charges)


#
class GraphAnalyzer:

    def __init__(self, coord, bo=None):

        self.nat = coord.shape[0]
        self.atoms = []
        self.coordinates = []
        for i in range(self.nat):
            self.atoms.append(coord.iloc[i, 0])
            self.coordinates.append(np.array(coord.iloc[i, 1:4]))
        self.wbo = bo

        # get topo matrix
        self.build_cn_matrix()
        # build up graph
        self.define_edge()
        self.degree = [deg for node, deg in self.Graph.degree()]

    def build_cn_matrix(self):
        self.Mcn = np.zeros((self.nat, self.nat), dtype=int)
        if self.wbo is not None:
            # build up cn matrix base on wbo file
            for i in range(self.wbo.shape[0]):
                at1, at2 = self.wbo.iloc[i, :2]
                at1 = int(at1) - 1
                at2 = int(at2) - 1
                self.Mcn[at1, at2] = self.Mcn[at2, at1] = 1
        else:
            from aRST.script.element import r0
            # build up cn matrix base on xyz itself
            for i in range(self.nat):
                for j in range(i + 1, self.nat):
                    dist = np.linalg.norm(self.coordinates[i] - self.coordinates[j])
                    threshold = r0[self.atoms[i]] + r0[self.atoms[j]]
                    if dist <= threshold * 1.1:
                        self.Mcn[i, j] = self.Mcn[j, i] = 1

    def define_edge(self):
        self.Graph = nx.Graph()
        for i in range(self.nat):
            for j in range(i):
                if self.Mcn[i, j]:
                    bond_length = np.linalg.norm(self.coordinates[i] - self.coordinates[j])
                    self.Graph.add_edge(i, j, weight=bond_length)

        G_sorted = nx.Graph()
        for node in sorted(self.Graph.nodes):
            G_sorted.add_node(node)
        for u, v, data in self.Graph.edges(data=True):
            G_sorted.add_edge(min(u, v), max(u, v), weight=data["weight"] if "weight" in data.keys() else 1)
        self.Graph = G_sorted

    def get_t2t_path(self, tarat1: int, tarat2: int):
        from itertools import islice
        # make sure tarat1 >= tarat2
        if tarat1 < tarat2:
            tarat1, tarat2 = tarat2, tarat1
        paths = nx.shortest_simple_paths(self.Graph, source=tarat1, target=tarat2, weight='weight')
        shortest_two = list(islice(paths, 2))

        if len(shortest_two) == 1:
            d = 5  # default as 5x covalent radii
            bond_length = self.Graph[tarat1][tarat2]['weight']
            return d * bond_length
        else:
            # Straighten the ring.
            length = sum(self.Graph[u][v]['weight'] for u, v in zip(shortest_two[1][:-1], shortest_two[1][1:])) * 0.9
            return length


def convert_for_json(d, exclude_keys=None):
    """
    Recursively convert a deep nested dict to be JSON serializable.
    - Exclude specific keys.
    - Convert float values to 4 decimal places.
    - Convert DataFrame to a JSON string.
    """
    if exclude_keys is None:
        exclude_keys = set()

    if isinstance(d, pd.DataFrame):
        return {"__dataframe__": d.to_json()}  # Store as JSON string
    elif isinstance(d, list):
        return [convert_for_json(v, exclude_keys) for v in d]
    elif isinstance(d, dict):
        return {k: convert_for_json(v, exclude_keys) for k, v in d.items() if k not in exclude_keys}
    elif isinstance(d, float):
        return round(d, 8)  # Round float values to 8 decimal places
    return d  # int, str, etc., remain unchanged


import json


class CompactJSONEncoder(json.JSONEncoder):
    def iterencode(self, obj, _one_shot=False):
        """Override default encoding to keep lists compact."""
        if isinstance(obj, list):
            return "[" + ",".join(map(self.encode, obj)) + "]"
        return super().iterencode(obj, _one_shot)


def write_rxnout():
    import aRST.script.initialization as storage
    from aRST.script.exploration.network import rxnnetwork
    import json
    data = {
        "nodes": [
            {
                "node": node,
                "info": rxnnetwork.get_reaction_info(node)}
            for node in rxnnetwork.graph.nodes
        ],
        "edges": list(rxnnetwork.graph.edges)
    }
    with open(f'{storage.setting.wd0}/reaction.json', "w") as file:
        file.write(json.dumps(data, indent=2, cls=CompactJSONEncoder))
        file.write('\n')


def write_strucout():
    import aRST.script.initialization as storage
    import json
    with open(f'{storage.setting.wd0}/struc.json', "w") as file:
        exclude_keys = ["atomic_connection"]
        file.write(json.dumps(convert_for_json(storage.allstruc.data, exclude_keys), indent=2))
        file.write('\n')



import os
from aRST.script.dictionaryInspector import DictionaryInspector
from aRST.script.toolbox import create_folder, write_xyz
from aRST.script.geom import alignment


def aligning(wd, STRUC_rec, setting_d, reactants):
    outputwd = create_folder(os.path.join(wd, "align"))
    struc_inspector = DictionaryInspector(STRUC_rec)
    setting_inspector = DictionaryInspector(setting_d)

    struc1, struc2 = reactants
    at1 = struc_inspector.get_value(f"{struc1}.struc_info.general.at")
    at2 = struc_inspector.get_value(f"{struc2}.struc_info.general.at")

    coord1 = struc_inspector.get_value(f"{struc1}.struc_info.general.coord")
    coord2 = struc_inspector.get_value(f"{struc2}.struc_info.general.coord")

    alignedcoord = alignment(coord1, coord2, at1, at2,
                             extra_setting={"maxovlap": setting_inspector.get_value(
                                 "searching_setting.inter.maxovlap", False),
                                 "aligning_distance": setting_inspector.get_value(
                                     "searching_setting.inter.aligning_distance", 5)},writingP=True)
    write_xyz(alignedcoord, f"aligned_coord")



import sys
from itertools import product
import numpy as np
import pandas as pd
import aRST.script.initialization as storage


def build_react_l(h_df, l_df, search_mode):
    """
    loop over Hdf and Ldf to build up atomic pairs
    """
    react_l = []
    for il, (istruc_l, row_l) in enumerate(l_df.items()):
        for ih, (istruc_h, row_h) in enumerate(h_df.items()):
            # Test search_mode
            if search_mode == 0 and il != ih:  # only search struc1 == struc2
                continue
            elif search_mode == 1 and il == ih:  # only search struc1 != struc2
                continue

            # loop row to get H and L values
            for iat_l, L in row_l.items():
                if isinstance(L, str) or np.isnan(L) or np.isinf(L):
                    continue
                for iat_h, H in row_h.items():
                    if isinstance(H, str) or np.isnan(H) or np.isinf(H):
                        continue
                    gap = np.round(np.round(L, 2) - np.round(H, 2), 2)
                    if int(istruc_l) <= int(istruc_h):
                        react_l.append({"struc_combi": [str(istruc_h), str(istruc_l)],
                                        "at_combi": [str(iat_h), str(iat_l)],
                                        "HOAO": H,
                                        "LUAO": L,
                                        "gap": gap})
                    else:
                        react_l.append({"struc_combi": [str(istruc_l), str(istruc_h)],
                                        "at_combi": [str(iat_l), str(iat_h)],
                                        "HOAO": H,
                                        "LUAO": L,
                                        "gap": gap})
    return react_l



def find_identical_at_from_afo(HOAO_a, HOAO_b, LUAO_a, LUAO_b, tol=0.01):
    out = {}

    # Process each structure that exists in HOAO_a
    for struc in HOAO_a.columns.unique():
        out[str(struc)] = []
        used_indices = set()

        for idx in HOAO_a.index:
            if idx in used_indices:
                continue

            # Get reference values
            ref_values = {
                'ha': HOAO_a.loc[idx, struc],
                'hb': HOAO_b.loc[idx, struc],
                'la': LUAO_a.loc[idx, struc],
                'lb': LUAO_b.loc[idx, struc]
            }

            # Ensure scalar values (handle Series cases)
            ref_values = {k: v.iloc[0] if isinstance(v, pd.Series) else v
                          for k, v in ref_values.items()}

            # Find degenerate atoms
            degenerate = [idx]
            used_indices.add(idx)

            # Compare with subsequent atoms
            for i in range(idx + 1, len(HOAO_a)):
                if i in used_indices:
                    continue

                # Get comparison values
                curr_values = {
                    'ha': HOAO_a.loc[i, struc],
                    'hb': HOAO_b.loc[i, struc],
                    'la': LUAO_a.loc[i, struc],
                    'lb': LUAO_b.loc[i, struc]
                }
                curr_values = {k: v.iloc[0] if isinstance(v, pd.Series) else v
                               for k, v in curr_values.items()}

                # Check tolerance
                if all(abs(curr_values[k] - ref_values[k]) < tol for k in ref_values):
                    degenerate.append(i)
                    used_indices.add(i)

            out[str(struc)].append({"atom": degenerate})

    return out


def refine_react_l(react_l, identical_atom_d, intra=False):
    # merge identical pais and keep the lowest gap
    # in the degenearted pairs, at can appear only once
    out_react_l = []
    used_combi = []
    if not intra:
        for item in react_l:
            struc1, struc2 = item["struc_combi"]  # istruc1 > istruc2
            at1, at2 = item["at_combi"]
            if tuple((int(struc1), int(struc2), int(at1), int(at2))) not in used_combi:
                # find all degenerated pair
                degenerate_at1 = [at1]
                degenerate_at2 = [at2]
                for tmp in identical_atom_d[struc1]:
                    if int(at1) in tmp["atom"]:
                        degenerate_at1 = tmp["atom"]
                        break
                for tmp in identical_atom_d[struc2]:
                    if int(at2) in tmp["atom"]:
                        degenerate_at2 = tmp["atom"]
                        break

                all_combi = list(product(degenerate_at1, degenerate_at2))
                if int(struc1) ==int(struc2):
                    all_combi2 = list(product(degenerate_at2, degenerate_at1))
                    all_combi.extend(all_combi2)
                used_l1 = []
                used_l2 = []
                pairs = []
                for combi in all_combi:
                    tat1, tat2 = combi
                    used_combi.append(tuple((int(struc1), int(struc2), int(tat1), int(tat2))))
                    if tat1 not in used_l1 and tat2 not in used_l2:
                        pairs.append([tat1, tat2])
                        used_l1.append(tat1)
                        used_l2.append(tat2)
                        # gap_value = next((tmpitem["gap"] for tmpitem in react_l if
                        #                   tmpitem["struc_combi"] == [str(struc1), str(struc2)]
                        #                   and tmpitem["at_combi"] == [str(tat1),str(tat2)]),None)

                item["at_combi"] = pairs

                if item not in out_react_l:
                    out_react_l.append(item)
            else:
                # change gap
                for searchitem in out_react_l:
                    search_struc1, search_struc2 = searchitem["struc_combi"]
                    if int(search_struc1) ==int(struc1) and int(search_struc2) ==int(struc2):
                        search_allcombi = searchitem["at_combi"]
                        triger = next((True for tmpitem in search_allcombi
                                       if int(tmpitem[0]) ==int(at1) and int(tmpitem[1]) ==int(at2)),False)
                        if triger:
                            currentgap = item["gap"]
                            oldgap = searchitem["gap"]
                            if currentgap < oldgap:
                                searchitem["gap"] = currentgap
                                searchitem["HOAO"] = item["HOAO"]
                                searchitem["LUAO"] = item["LUAO"]
                                break
    else:
        # todo
        for item in react_l:
            struc = item["struc_combi"][0]
            at1, at2 = item["at_combi"]
            if tuple((int(struc), int(at1), int(at2))) not in used_combi:
                # find all degenerated pair
                degenerate_at1 = [at1]
                degenerate_at2 = [at2]
                for tmp in identical_atom_d[struc]:
                    if at1 in tmp["atom"]:
                        degenerate_at1 = tmp["atom"]
                        break
                for tmp in identical_atom_d[struc]:
                    if at2 in tmp["atom"]:
                        degenerate_at2 = tmp["atom"]
                        break
                all_combi = list(product(degenerate_at1, degenerate_at2))
                for combi in all_combi:
                    tat1, tat2 = combi
                    used_combi.append(tuple((int(struc), int(tat1), int(tat2))))
                used_l = []
                pairs = []
                for pair in all_combi:
                    at1, at2 = pair
                    if at1 not in used_l and at2 not in used_l:
                        pairs.append([at1, at2])
                        used_l.append(at1)
                        used_l.append(at2)
                item["at_combi"] = pairs
                if item not in out_react_l:
                    out_react_l.append(item)
            else:
                # change gap
                for searchitem in out_react_l:
                    search_struc = searchitem["struc_combi"][0]
                    if search_struc == struc:
                        search_allcombi = searchitem["at_combi"]
                        if tuple([at1, at2]) in search_allcombi:
                            currentgap = item["gap"]
                            oldgap = searchitem["gap"]
                            if currentgap < oldgap:
                                searchitem["gap"] = currentgap
                                searchitem["HOAO"] = item["HOAO"]
                                searchitem["LUAO"] = item["LUAO"]
                                break
        # change key name struc_combi to struc
        for item in out_react_l:
            item["struc"] = item["struc_combi"][0]
            del item["struc_combi"]
            # make sure iat1>iat2
            pairs = [[max(at1, at2), min(at1, at2)] for at1, at2 in item["at_combi"]]
            item["at_combi"] = pairs
    return out_react_l


class SPR:
    # atom index starts with 0
    def __init__(self, reactants,
                 allow_hydrogen=True,
                 allow_halogen=True,
                 afo_l=None):
        self.reactants = reactants
        self.allowhydrogen = allow_halogen
        self.allowhalogen = allow_halogen

        if afo_l is None:
            maxnat = max(storage.allstruc.get_value(f'{i}.struc_info.reactivity.afo').shape[0] for i in self.reactants)
        else:
            maxnat = max(i.shape[0] for i in afo_l)

        # collecting HL values into dataframe
        self.HOAO_a = pd.DataFrame(columns=self.reactants, index=range(maxnat))
        self.HOAO_b = pd.DataFrame(columns=self.reactants, index=range(maxnat))
        self.LUAO_a = pd.DataFrame(columns=self.reactants, index=range(maxnat))
        self.LUAO_b = pd.DataFrame(columns=self.reactants, index=range(maxnat))

        newcolumn_names = ["HOAO_a (eV)", "HOAO_b (eV)", "LUAO_a (eV)", "LUAO_b (eV)"]
        dfs = [self.HOAO_a, self.HOAO_b, self.LUAO_a, self.LUAO_b]

        for istruc in range(len(self.reactants)):
            strucid = self.reactants[istruc]
            if afo_l is None:
                mldf_original = storage.allstruc.get_value(f'{strucid}.struc_info.reactivity.afo').copy()

            else:
                mldf_original = afo_l[istruc].copy()
                # remove all hydrogen/halogen if not allowed
            if not allow_hydrogen:
                # get index
                hindex = mldf_original[mldf_original['Atom'] == 1].index.tolist()
                mldf_original.loc[hindex, "HOAO_a (eV)"] = -np.inf
                mldf_original.loc[hindex, "HOAO_b (eV)"] = -np.inf
                mldf_original.loc[hindex, "LUAO_a (eV)"] = np.inf
                mldf_original.loc[hindex, "LUAO_a (eV)"] = np.inf

            if not allow_halogen:
                # get index
                hindex = mldf_original[mldf_original['Atom'] == 9 or
                                       mldf_original['Atom'] == 17 or
                                       mldf_original['Atom'] == 35 or
                                       mldf_original['Atom'] == 53 or
                                       mldf_original['Atom'] == 85].index.tolist()
                # change values
                mldf_original.loc[hindex, "HOAO_a (eV)"] = -np.inf
                mldf_original.loc[hindex, "HOAO_b (eV)"] = -np.inf
                mldf_original.loc[hindex, "LUAO_a (eV)"] = np.inf
                mldf_original.loc[hindex, "LUAO_a (eV)"] = np.inf

            for df, column_name in zip(dfs, newcolumn_names):
                tmpl = list(mldf_original.loc[:, column_name])
                for j in range(maxnat - len(tmpl)):
                    tmpl.append(np.nan)
                df.iloc[:, istruc] = tmpl  # .reset_index(drop=True)

    def inter(self, spr_mode):
        identical_atom_d = find_identical_at_from_afo(self.HOAO_a,self.HOAO_b,self.LUAO_a,self.LUAO_b)
        final_react = []
        if spr_mode == 1:
            react_aa = build_react_l(self.HOAO_a, self.LUAO_a, search_mode=1)
            react_bb = build_react_l(self.HOAO_b, self.LUAO_b, search_mode=1)
            for react in [react_aa, react_bb]:
                final_react.extend(react)
        elif spr_mode == 2:
            react_ab = build_react_l(self.HOAO_a, self.LUAO_b, search_mode=1)
            react_ba = build_react_l(self.HOAO_b, self.LUAO_a, search_mode=1)
            for react in [react_ab, react_ba]:
                final_react.extend(react)

        else:
            react_aa = build_react_l(self.HOAO_a, self.LUAO_a, search_mode=1)
            react_bb = build_react_l(self.HOAO_b, self.LUAO_b, search_mode=1)
            react_ab = build_react_l(self.HOAO_a, self.LUAO_b, search_mode=1)
            react_ba = build_react_l(self.HOAO_b, self.LUAO_a, search_mode=1)

            for react in [react_aa, react_bb, react_ab, react_ba]:
                final_react.extend(react)

        final_react = sorted(refine_react_l(final_react, identical_atom_d), key=lambda x: x['gap'])
        return final_react

    def intra(self, spr_mode):
        identical_atom_d = find_identical_at_from_afo(self.HOAO_a,self.HOAO_b,self.LUAO_a,self.LUAO_b)
        final_react = []
        if spr_mode == 1:
            react_aa = build_react_l(self.HOAO_a, self.LUAO_a, search_mode=0)
            react_bb = build_react_l(self.HOAO_b, self.LUAO_b, search_mode=0)
            for react in [react_aa, react_bb]:
                final_react.extend(react)
        elif spr_mode == 2:
            react_ab = build_react_l(self.HOAO_a, self.LUAO_b, search_mode=0)
            react_ba = build_react_l(self.HOAO_b, self.LUAO_a, search_mode=0)
            for react in [react_ab, react_ba]:
                final_react.extend(react)

        else:
            react_aa = build_react_l(self.HOAO_a, self.LUAO_a, search_mode=0)
            react_bb = build_react_l(self.HOAO_b, self.LUAO_b, search_mode=0)
            react_ab = build_react_l(self.HOAO_a, self.LUAO_b, search_mode=0)
            react_ba = build_react_l(self.HOAO_b, self.LUAO_a, search_mode=0)

            for react in [react_aa, react_bb, react_ab, react_ba]:
                final_react.extend(react)

        tmp_react = sorted(refine_react_l(final_react, identical_atom_d, intra=True), key=lambda x: x['gap'])
        # del all connected bond
        allbonds = SDP(self.reactants, self.allowhydrogen, self.allowhalogen).allbonds
        final_react = []
        for item in tmp_react:
            struc = item["struc"]
            atl = item["at_combi"][0]
            trigger = True
            for item2 in allbonds:
                if item2["struc"] == struc:
                    if item2["at_combi"] == atl or len(set(atl))==1:
                        trigger=False
            if trigger:
                final_react.append(item)
        return final_react


class SDP:
    # atom index starts with 0
    def __init__(self, reactants,
                 allow_hydrogen=True,
                 allow_halogen=True,
                 bo_l=None, coord_l=None):
        self.reactants = reactants
        self.allbonds = []
        self.dfs = []
        if not bo_l is None:
            self.dfs = bo_l
            self.coord_l = coord_l
        else:
            self.dfs = [storage.allstruc.get_value(f"{_}.struc_info.reactivity.bo") for _ in self.reactants]
            self.coord_l = [storage.allstruc.get_value(f"{_}.struc_info.general.coord") for _ in self.reactants]

        for i in range(len(reactants)):
            struc = self.reactants[i]
            coord = self.coord_l[i]
            bodf = self.dfs[i]
            if bodf is not None:
                # read through df
                for index in range(bodf.shape[0]):
                    at1, at2 = bodf.iloc[index, :2]
                    at1 = int(at1) - 1
                    at2 = int(at2) - 1
                    ele1, ele2 = coord.iloc[at1, 0], coord.iloc[at2, 0]
                    if not allow_hydrogen:
                        if ele1 == "H" or ele2 == "H":
                            continue
                    if not allow_halogen:
                        if ele1 in ["F", "Cl", "cl", "Br", "I"] or ele2 in ["F", "Cl", "cl", "Br", "I"]:
                            continue
                    mbo = bodf.iloc[index, 2]
                    if mbo < 0.3:
                        continue
                    if at1 < at2:
                        at1, at2 = at2,at1
                    else:
                        at1, at2 = at1, at2
                    if bodf.shape[1] > 2:
                        embo = bodf.iloc[index, 3]
                        self.allbonds.append({"struc": struc,
                                              "at_combi": [str(at1), str(at2)],
                                              "mbo": mbo,
                                              "embo": embo})
                    else:
                        self.allbonds.append({"struc": struc,
                                              "at_combi": [str(at1), str(at2)],
                                              "mbo": mbo,
                                              })

    def build_react_l(self, mode, tol=0.01,gaplim=0.9):
        # refine react_l
        react_l = []
        for item in self.allbonds:
            struc = item["struc"]
            at1, at2 = item["at_combi"]
            mbo = item["mbo"]
            if mode == 0 and mbo>gaplim:
                continue
            if not react_l:
                react_l.append(item)
                react_l[0]["at_combi"] = [[at1, at2]]
            else:
                # loop over react_l if degenerate dissociation pair
                degenerate = False
                for item2 in react_l:
                    if int(item2["struc"]) == int(struc) and abs(mbo - item2["mbo"]) < tol:
                        degenerate = True
                        if not any([at1 in _ for _ in item2["at_combi"]]) and not any(
                                [at2 in _ for _ in item2["at_combi"]]):
                            item2["at_combi"].append([at1, at2])
                            break
                if not degenerate:
                    react_l.append(item)
                    react_l[-1]["at_combi"] = [[at1, at2]]

        if mode == 0:
            return sorted(react_l, key=lambda x: x["mbo"])
        else:
            # embo as bo
            if "embo" in react_l[0].keys():
                return sorted(react_l, key=lambda x: x["embo"])
            else:
                print("Warning: embo doesn't exist in struc info, change back to MBO assessment")
                return sorted(react_l, key=lambda x: x["mbo"])



import sys
from aRST.script.toolbox import mydict, get_all_keys
import os
import shutil

required_input = {
    "read": {
        "general": {
            "dftopt": (bool, False),  # doing dftopt when first reading input coord
            "gfn2opt": (bool, False),  # doing gfn2 when first reading input coord
            "jumprefine": (bool, False),  # direct read struc info from folder struc/0,1,2,.....
        }
    },
    "searching_setting": {
        "general": {
            "cycle": (int, 1),  # reaction simulation cycle
            "elim": ((int, float), 300),
            # energy limitation to kill this reaction path in current path, unit in kcal/mol
            "hydrogen": (bool, True),  # include hydrogen element in the AFO searching
            "halogen": (bool, True),  # include halogen elements in the AFO searching
            "spr_mode": (int, 0),
            # AFO gap formation rules: 0: take both spin same/oppo combinations into account; 1: only same spin; 2: only oppo spin
            "scan_mode": (int, 0),  # bond formation method: 0: using xtb scan; 1: using harmonic FF
            "sep_mode": (int, 0),
            # separation rules for preliminary product structure: 0: sep fragments before gfn2 opt; 1: opt product before separation
            "mb_level": (str, None),  # "loose" to only screen with topology spectrum
            "keepmolden": (bool, False)  # del all molden files to avoid large I/O writing
        },
        "reactivity": {
            "afo_method": (str, "gfn2"),  # method for afo feature calculations
        },
        "inter": {
            "number": (int, 0),  # number of intermolecular reaction paths to be searched in one cycle
            "search_mode": (int, 1),
            # searching rules: 0: only perform bonding simulation to same structure (A+A); 1: ..to different structure (A+B); 2: both (A+A and A+B)
            "at1": (int, None),
            # manually define the search pair, index strarts from 0. at1,at3 belong to struc 0; at2,at4 belong to struc1
            "at2": (int, None),
            "at3": (int, None),
            "at4": (int, None),
            "loose_bond": (bool, False),  # loose a bit the weakest bond in the structure after traget bond formation
            "maxovlap": (bool, False),  # alignment method for achiving maximum overlap of reacting structures
            "aligning_distance": ((int, float), 5),  # scale of aligning distance of reacting structures
            "concerted_pair": (bool, False)  # reate concerted pair in simulations (good for DA reactions)
        },
        "intra": {
            "number": (int, 0),  # number of intramolecular reaction paths to be searched in one cycle
            "at1": (int, None),
            # manually define the search pair, index strarts from 0. at1,at3 belong to struc 0; at2,at4 belong to struc1
            "at2": (int, None),
            "at3": (int, None),
            "at4": (int, None),
            "concerted_pair": (bool, False)  # reate concerted pair in simulations (good for DA reactions)
        },
        "disso": {
            "number": (int, 0),  # number of dissociative reaction paths to be searched in one cycle
            "at1": (int, None),
            # manually define the search pair, index strarts from 0. at1,at3 belong to struc 0; at2,at4 belong to struc1
            "at2": (int, None),
            "at3": (int, None),
            "at4": (int, None),
            "mode": (int, 0),  # 0: use mbo as bo; 1: use embo as bo for ranking bond term strength
            "concerted_pair": (bool, False),  # reate concerted pair in simulations (good for DA reactions)
            "pulling_distance": ((int, float, str), "auto"),  # scale of pulling distance
            "depotonation": (bool, False),  # checking depotonated reactant before dissociation
            "potonation": (bool, False),  # checking potonated reactant before dissociation
        }
    },
    "xtb_setting": {  # this settings go into both gfn sp and opt calculations
        "method": (str, "gfn2"),  # GFN lvl
        "optlevel": (str, None),  # xtb acceptable optimization levels
        "alpb": (str, "ch2cl2"),  # ALPB solvation model, aRST only pass keyword to xtb program
        "struc_etemp": (int, 300),  # electronic temp for structure opt, recommended using RT temp.
    },
    "orca_setting": {  # this settings go into both DFT sp and opt calculations
        "functional": (str, "r2scan-3c"),  # functional method
        "basis": (str, None),  # basis set
        "dftopt": (bool, False),  # if doing DFT opt for product structure
        "solmode": (str, "ch2cl2"),  # CPCM solvation model, aRST only pass keyword to orca program
        "ncpu": (int, 1),  # parallel calculations for orca
    }
}

condition_md = [
    {
        "condition": [("searching_setting", "general", "MD", True)],
        "required_block": {
            "searching_setting": {
                "MD": (bool, False),  # perform short md simulations to test the stability of product structure
                "md": {
                    "time": (int, 10),  # total run time of MD simulation, unit in ps
                    "step": (int, 1),  # time step for propagation
                    "temp": (int, 300),  # electronic temp, by defalut using scan temp, unit in K
                    "shake": (int, 1),  # 1: constrain H bonds; 2: constrain all bonds
                    "dump": (int, 50),  # interval for trajectory printout
                }
            }
        },
        "message": "Since MD simulation is asked, the block [searching_setting.md] must be provided."
    }
]
condition_AFO_orca = [
    {
        "condition": [("searching_setting", "reactivity", "afo_method", "stda"),
                      ("searching_setting", "reactivity", "MO_method", "orca")],
        "required_block": {
            "searching_setting": {
                "reactivity": {
                    "afo_method": (str, "stda"),
                    "MO_method": (str, "orca"),
                    "afo_orca_functional": (str, "bhlyp"),  # functional method
                    "afo_orca_basis": (str, None),  # basis set
                    "afo_orca_cpcm": (str, "ch2cl2"),  # CPCM solvation model, aRST only pass keyword to orca program
                }
            }
        },
        "message": "Since AFO features are asked to calculated at DFT methods in ORCA, the DFT methods must be provided."
    }
]
condition_AFO_terachem = [
    {
        "condition": [("searching_setting", "reactivity", "afo_method", "stda"),
                      ("searching_setting", "reactivity", "MO_method", "terachem")],
        "required_block": {
            "searching_setting": {
                "reactivity": {
                    "afo_method": (str, "stda"),
                    "MO_method": (str, "terachem"),
                    "afo_orca_functional": (str, "bhlyp"),  # functional method
                    "afo_orca_basis": (str, None),  # basis set
                    "afo_tc_epsilon": ((float, int), 9.08),
                    # CPCM solvation model, aRST only pass keyword to Terachem program
                }
            }
        },
        "message": "Since AFO features are asked to calculated at DFT methods in Terachem, the DFT methods must be provided."
    }
]
condition_xtbscan = [
    {
        "condition": [("searching_setting", "general", "scan_mode", 0)],
        "required_block": {
            "xtb_setting": {
                "scan_etemp": (int, 1000),  # electronic temp for xtbscan/harmonic FF, recommended using higher temp.
                "k": (int, 1),  # geom constrain strength
                "step": (int, 20),  # pushing/pulling steps
            }
        },
        "message": ""
    }
]
condition_FF = [
    {
        "condition": [("searching_setting", "general", "scan_mode", 1)],
        "required_block": {
            "xtb_setting": {
                "scan_etemp": (int, 1000),  # electronic temp for xtbscan/harmonic FF, recommended using higher temp.
                "k": (int, 1),  # geom constrain strength
            }
        },
        "message": ""
    }
]

conditions = []
for i in (condition_FF, condition_xtbscan, condition_md, condition_AFO_terachem, condition_AFO_orca):
    conditions.append(i)


def checking_parameter(setting):
    assert isinstance(setting, dict) and setting
    schema = mydict(required_input)
    # config = mydict(setting)
    for item in conditions:
        item = item[0]
        condition = item["condition"]
        add = True
        for item2 in condition:
            key = ".".join(list(item2[:-1]))
            expect_val = item2[-1]
            if isinstance(expect_val, str):
                expect_val = expect_val.lower()
            if setting.get_value(key) != expect_val:
                add = False
                break
        if add:
            for block_key, block_value in item["required_block"].items():
                if block_key not in required_input:
                    required_input[block_key] = {}
                for subkey, subval in block_value.items():
                    required_input[block_key][subkey] = subval

    all_keys_in_config = get_all_keys(setting)
    all_keys_in_shema = get_all_keys(schema)
    allcorrect = True
    for item in all_keys_in_config:
        # check if input key exists
        if item in all_keys_in_shema:
            # check if type is correct
            val = setting.get_value(item)
            expected_type = schema.get_value(item)[0]
            if isinstance(expected_type, (list, tuple)):
                accepted_type = list(expected_type)
            else:
                accepted_type = [expected_type]

            if type(val) in accepted_type:
                pass
            else:
                allcorrect = False
                print("Error type in", val, ": This type is not accepted for", item, f"{accepted_type}")
        elif "geom" in item:
            # geom block will be checked in init
            continue
        elif "wd" in item:
            # accepted in aRST 1.0
            continue
        else:
            allcorrect = False
            print("Error parameter in ", item, ": This parameter is not accept in aRST")
    if not allcorrect:
        raise ValueError("wrong input!")
    else:
        # adding missing parameter to default val
        missing_par = [x for x in all_keys_in_shema if x not in all_keys_in_config]
        print("The following missing parameters will be set to their default values:")
        for par in missing_par:
            defalut_val = schema.get_value(par)[1]
            if defalut_val is not None and not (isinstance(defalut_val, bool) and not defalut_val):
                setting.set_value(par, defalut_val)
                print(f"{par} -> {defalut_val}")
        print()
    return setting



def checking_env(setting):
    required_vars = ["ORCA_BIN","XTB_BIN"]#,"TRVT_BIN"
    if setting.get_value("searching_setting.reactivity.afo_method") =="stda":
        required_vars.append("STDA_BIN")
        required_vars.append("MWFN_BIN")
        if setting.get_value("searching_setting.reactivity.MO_method") =="terachem":
            required_vars.append("TC_BIN")
    elif setting.get_value("searching_setting.reactivity.afo_method") =="gfn2":
        required_vars.append("XTBML_BIN")
    if setting.get_value("searching_setting.disso.depotonation") or setting.get_value("searching_setting.disso.potonation"):
        required_vars.append("CREST_BIN")

    missing = [var for var in required_vars if var not in os.environ]
    if missing:
        print("Missing environment variables:", ", ".join(missing))
        print("Please set them using 'export VAR=VALUE' in your shell.")
        exit(1)




import os

import numpy as np
import pandas as pd
from aRST.script.element import r0,element,scaled_r0
import warnings
warnings.filterwarnings( "ignore" )


class Coord:
    def __init__(self, coord):
        self.coord = coord

    def get_principal_moment_vec(self, at):
        t1, t2, t3, t4, t5, t6 = 0, 0, 0, 0, 0, 0
        # matrix for moments of inertia is of form, with CN in mass as weight
        # *           |   y**2+z**2                         |
        # *           |    -y*x       z**2+x**2             | -i =0
        # *           |    -z*x        -z*y       x**2+y**2 |
        xv, yv, zv = 0, 0, 0
        neighbor_atl = []
        for i in range(self.coord.shape[0]):
            if i != at:
                x = self.coord.iloc[i, 1] - self.coord.iloc[at, 1]
                y = self.coord.iloc[i, 2] - self.coord.iloc[at, 2]
                z = self.coord.iloc[i, 3] - self.coord.iloc[at, 3]
                r = np.linalg.norm(np.array([x, y, z]))
                rcov_c = float(scaled_r0[f'{self.coord.iloc[at, 0]}'])
                rcov_i = float(scaled_r0[f'{self.coord.iloc[i, 0]}'])
                rr = (rcov_i + rcov_c) / r
                # add connected at's xyz
                if rr > 0.8:
                    neighbor_atl.append([self.coord.iloc[i, 1], self.coord.iloc[i, 2], self.coord.iloc[i, 3]])
                # add up I
                cn = 1 / (1 + np.exp(- 16 *(rr - 1)))
                atmass = 0.5 * cn / (r ** 2)
                t1 = t1 + atmass * (y ** 2 + z ** 2)
                t2 = t2 + atmass * (-y * x)
                t3 = t3 + atmass * (z ** 2 + x ** 2)
                t4 = t4 + atmass * (-z * x)
                t5 = t5 + atmass * (-z * y)
                t6 = t6 + atmass * (x ** 2 + y ** 2)
                # add up molecular orientation
                xv = xv + element[f'{self.coord.iloc[i, 0]}']* x
                yv = yv + element[f'{self.coord.iloc[i, 0]}'] * y
                zv = zv + element[f'{self.coord.iloc[i, 0]}'] * z
        # print(cn)
        mv = np.array([xv, yv, zv])
        arep = np.zeros((3, 3))
        arep[0][0], arep[1][0], arep[1][1], arep[2][0], arep[2][1], arep[2][2] = t1, t2, t3, t4, t5, t6
        arep = arep + arep.T - np.diag(arep.diagonal())

        v, P = np.linalg.eigh(arep)
        #     print(v.round(3))
        #     print(P.round(3))
        return mv, P, v, neighbor_atl

    def get_punkt_coord(self, at):
        a = np.array([self.coord.iloc[at, 1],
                      self.coord.iloc[at, 2],
                      self.coord.iloc[at, 3]])
        return a

    def get_nat(self):
        return self.coord.shape[0]

    def shift(self, at, d):
        # shift the molecule with aktiv as center to (0,0,d)
        x = self.coord.iloc[at, 1]
        y = self.coord.iloc[at, 2]
        z = self.coord.iloc[at, 3] - d
        for i in range(self.coord.shape[0]):
            self.coord.iloc[i, 1] = self.coord.iloc[i, 1] - x
            self.coord.iloc[i, 2] = self.coord.iloc[i, 2] - y
            self.coord.iloc[i, 3] = self.coord.iloc[i, 3] - z
        return self.coord

    def change_cs(self, P):
        for i in range(self.coord.shape[0]):
            x, y, z = self.coord.iloc[i, 1], self.coord.iloc[i, 2], self.coord.iloc[i, 3]
            a = [x, y, z]
            a_new = np.dot(P.T, a)
            self.coord.iloc[i, 1], self.coord.iloc[i, 2], self.coord.iloc[i, 3] = a_new[0], a_new[1], a_new[2]
        return self.coord

    def write_xyz(self, name):
        xyz = self.coord.round(5)
        nat = xyz.shape[0]
        with open(f'{name}.xyz', 'w') as f:
            f.write(f'{nat}\n')
            f.write('\n')
            for i in range(nat):
                at, x, y, z = xyz.iloc[i, 0], xyz.iloc[i, 1], xyz.iloc[i, 2], xyz.iloc[i, 3]
                f.write('%2s %10.5f %10.5f %10.5f\n' % (at, x, y, z))
            f.write('\n')

    def write_p(self, at, P, name, dst, length=2, porder=None):
        # display p vector in chimera, defalut length =2
        a = Coord.get_punkt_coord(self, at)
        x, y, z = a[0], a[1], a[2]
        P_scalred = np.copy(P)
        for i in range(P_scalred.shape[1]):
            tmp = np.linalg.norm(P[:, i])
            P_scalred[:, i] = P_scalred[:, i] * length / tmp

        targetfile = os.path.join(dst, f'{name}.bild')
        color0 = ["green", "blue", "red"]
        if porder is not None:
            color = [color0[i] for i in porder]
        else:
            color = color0
        with open(targetfile, 'w+') as f2:
            # Coordinante system
            f2.write('.color black\n')
            f2.write(f'.arrow 0 0 0 0 0 6 0.05 0.150000\n')
            f2.write(f'.arrow 0 0 0 6 0 0 0.05 0.150000\n')
            f2.write(f'.arrow 0 0 0 0 6 0 0.05 0.150000\n')

            f2.write(f'.color {color[2]}\n')
            f2.write(
                f'.arrow {x} {y} {z} {x + P_scalred[:, 2][0]} {y + P_scalred[:, 2][1]} {z + P_scalred[:, 2][2]} 0.05 0.150000\n')

            f2.write(f'.color {color[0]}\n')
            f2.write(
                f'.arrow {x} {y} {z} {x + P_scalred[:, 0][0]} {y + P_scalred[:, 0][1]} {z + P_scalred[:, 0][2]} 0.05 0.150000\n')

            f2.write(f'.color {color[1]}\n')
            f2.write(
                f'.arrow {x} {y} {z} {x + P_scalred[:, 1][0]} {y + P_scalred[:, 1][1]} {z + P_scalred[:, 1][2]} 0.05 0.150000\n')
        return

    def get_dihedral_angle(self, at1, at2, at3, at4):
        # struc1: at1, at3 and at1 = 0,0,0
        # struc2: at2,at4 and at2 = 0,0,z
        a = self.get_punkt_coord(at1)
        b = self.get_punkt_coord(at2)
        c = self.get_punkt_coord(at3)
        d = self.get_punkt_coord(at4)

        ra = b - a
        rb = c - b
        rc = d - c

        n1 = np.cross(ra, rb)
        n2 = np.cross(rb, rc)

        n1_normalized = n1 / np.linalg.norm(n1)
        n2_normalized = n2 / np.linalg.norm(n2)

        # Calculate the dihedral angle in radians
        # sign = np.sign(np.dot(np.cross(ra, rb), rc))
        theta = np.arccos(np.dot(n1_normalized, n2_normalized))

        #         # # Determine the sign of the angle using the dot product between n1 and rc
        if np.dot(n1, rc) < 0.0:
            theta = -theta

        # Convert the angle to degrees
        theta = np.degrees(theta)

        return theta

    def test_if_cross(self, at1, at2, at3, at4):
        # atl =  [[at1,at2],[at3,at4]]
        # at1 = 0,0,0; at2 = 0,0,z
        d1 = self.get_dihedral_angle(at3, at1, at2, at4)
        d2 = self.get_dihedral_angle(at3, at1, at4, at2)

        return abs(d1) >= abs(d2)

    def get_ats_distance(self, at1, at2):
        a = self.get_punkt_coord(at1)
        b = self.get_punkt_coord(at2)
        c = b - a
        return np.linalg.norm(c)

    def get_rr_distance(self, at):
        ele = self.coord.iloc[at, 0]
        return r0[ele]


def get_angle(v1, v2): # 0-180
    r1 = np.round(np.linalg.norm(v1), 1)
    r2 = np.round(np.linalg.norm(v2), 1)
    dot_product = np.round(np.dot(v1, v2), 1)
    try:
        theta = np.arccos(dot_product / (r1 * r2)) * 180 / np.pi
    except RuntimeWarning:
        theta = 90
    return theta


def check_orient(P, mv, rh_cs=True):
    # default to optput right-handed P
    tmp, tmp2 = [], []
    current_type = check_rl(P)  # true = left handed

    for i in range(3):
        p = np.array(P[:, i])
        theta = get_angle(p, mv)
        tmp2.append(abs(abs(theta) - 90))

        if theta < 90:
            tmp.append(i)
            P[:, i] = -p

    if len(tmp) == 1 or len(tmp) == 3:
        if (rh_cs and current_type) or (not rh_cs and not current_type):
            i = tmp2.index(min(tmp2))
            P[:, i] = -P[:, i]
        else:
            return P

    return P


def check_range(P, v):
    #
    v = v.round(2)
    v = [i for i in v]
    # a=b=c spherical pr=pc
    # a<b=c prolate pr=pa !
    # a=b<c oblate pr=pc
    # a=0 b=c linear pr=pc
    # a+b = c oblate, plane pr=pc
    column_order = [0, 1, 2]
    if v[1] == v[2] and v[0] != 0:  # prolate
        column_order = [0, 1, 2]
        pr_index = v.index(min(v))
        for i in range(2 - pr_index):
            column_order.insert(0, column_order.pop())
        P = P[:, column_order]
    else:
        P = P
    return P,column_order


def check_rl(P):
    # retrun true if P is left-handed
    a, b, c = P[::, 0], P[::, 1], P[::, 2]
    cross_p = [np.round(j, 3) for j in np.cross(b, c)]
    dot_p = np.dot(a, cross_p)
    if dot_p > 0:
        return True
    else:
        return False


def l2r(P):
    column_order = [1, 0, 2]
    P = P[:, column_order]
    return P


def alignment(coord1, coord2, at1, at2, extra_setting=None,writingP=False):
    # copy coord so that won't affect global variant
    stuc1 = coord1.copy(deep=True)
    stuc2 = coord2.copy(deep=True)

    stuc1 = Coord(stuc1).shift(d=0, at=at1)
    mv1, P1, v1, neighbor_atl1 = Coord(stuc1).get_principal_moment_vec(at1)

    # make sure p point into the direction of the smallest resistance and keep right-handed
    P1 = check_orient(P1, mv1)
    if writingP:
        Coord(stuc1).write_xyz("coord1")
        Coord(stuc1).write_p(at1, P1, "p1_primitive", os.getcwd(), 2)

    # find main reaction direction
    P1,new_porder = check_range(P1,v1)

    # change coordinante system to main axes system
    stuc1 = Coord(stuc1).change_cs(P1)
    xyz1 = stuc1.copy(deep=True)
    if writingP:
        Coord(stuc1).write_p(at1, P1, "p1_oriented", os.getcwd(), 2, new_porder)

    # now do 2. structure
    stuc2 = Coord(stuc2).shift(d=0, at=at2)
    mv2, P2, v2, neighbor_atl2 = Coord(stuc2).get_principal_moment_vec(at2)

    P2 = check_orient(P2, mv2)
    if writingP:
        Coord(stuc2).write_xyz("coord2")
        Coord(stuc2).write_p(at2, P2, "p2_primitive", os.getcwd(), 2)


    # find main reaction direction
    P2,new_porder = check_range(P2,v2)

    P2[:, 2] = -P2[:, 2]

    P2[:, 0] = -P2[:, 0]
    P2[:, 1] = -P2[:, 1]

    maxovlap = extra_setting.get("maxovlap",False)
    if maxovlap:
        P2[:, 0] = -P2[:, 0]
        P2[:, 1] = -P2[:, 1]

    stuc2 = Coord(stuc2).change_cs(P2)
    if writingP:
        Coord(stuc2).write_p(at2, P2, "p2_oriented", os.getcwd(), 2, new_porder)

    # merge two structures
    name1 = xyz1.iloc[at1, 0]
    name2 = stuc2.iloc[at2, 0]
    rr = (r0[name1] + r0[name2])
    relativ_r = extra_setting.get('aligning_distance', 5)  # 5 times of rr by default
    d = np.round(rr * relativ_r, 2)

    stuc2 = Coord(stuc2).shift(at2, d)
    xyz2 = stuc2.copy(deep=True)  # for the bug if cood1=cood2, the class will be inherited

    coord_merge = pd.concat([xyz1, xyz2], ignore_index=True)
    return coord_merge



import pandas as pd
from aRST.script.geom import Coord



def build_push_l(coord, at_l, nat1=0,d=0.98):
    # avoiding crossing push atomic pairs
    if len(at_l) ==2:
        at1 = int(at_l[0][0])
        at2 = int(at_l[0][1]) + int(nat1)
        at3 = int(at_l[1][0])
        at4 = int(at_l[1][1]) + int(nat1)
        if Coord(coord).test_if_cross(at1, at2, at3, at4):
            current_atl = [[at1, at4], [at2, at3]]
        else:
            current_atl = [[at1, at2], [at3, at4]]
        # current_atl = [[at1, at2]]
    else:
        current_atl = [[int(at_l[0][0]), int(at_l[0][1]) + nat1]]

    ## at_l: index starting with 0; push_l: index starting with 1
    push_l = [(k[0] + 1,  # at1 (index starts with 1)
               k[1] + 1,  # at2 (index starts with 1)
               Coord(coord).get_ats_distance(k[0], k[1]),  # d_12
               (Coord(coord).get_rr_distance(k[0]) +
                   Coord(coord).get_rr_distance(k[1])*d)) for k in current_atl]  # rr_12

    return push_l

def get_endxyz_from_mb(constrain_l, coordinates, ele,set_angles=True,custum_constraints=None):
    from molbar.idealize import idealize_structure_from_coordinates
    # Define constraints
    if custum_constraints is None:
        extra_constraints = {
            'bond_order_assignment': True,
            'cycle_detection': True,
            'set_edges': False,
            'set_angles': set_angles,
            'set_dihedrals': False,
            'set_repulsion': True,
            'repulsion_charge': 100.0,
            'constraints': {'bonds': constrain_l}
        }
    else:
        extra_constraints = custum_constraints

    # Call the idealization function
    res = idealize_structure_from_coordinates(coordinates, ele, input_constraint=extra_constraints)
    coord_l, ele = res[1], res[2]

    # Format results into a dataframe
    coord = pd.DataFrame(coord_l, columns=['X', 'Y', 'Z'])
    coord.insert(0, 'Symbol', ele)
    coord.columns = [''] * len(coord.columns)
    return coord



import traceback
from os.path import join, exists
import os
import shutil
import subprocess
import re
from aRST.script.toolbox import mydict
import pandas as pd


def remove_file(old_path, new_path):
    filelist = os.listdir(old_path)

    for file in filelist:
        src = join(old_path, file)
        dst = join(new_path, file)
        shutil.move(src, dst)


class CallORCA:
    @classmethod
    def default_settings(cls, setting):
        """Dynamically generates default settings based on the `setting` class."""
        return {
            "functional": setting.get_value("orca_setting.functional", "r2scan-3c"),
            "basis": setting.get_value("orca_setting.basis", " "),
            "charge": 0,
            "m": 1,
            "fn": "coord",
            "caltyp": "SP",
            "caltyp2": "RKS",  # This can be updated dynamically if needed
            "solmode": setting.get_value("orca_setting.solmode", ""),
            "ncpu": setting.get_value("orca_setting.ncpu", 1),
        }

    def __init__(self, wd, command_d=None, custom_settings=None):
        self.wd = wd
        if command_d is not None:
            self.cmd = mydict(command_d)
        else:
            import aRST.script.initialization as storage
            command_d = self.default_settings(storage.setting)
            if custom_settings is not None:
                command_d.update(custom_settings)

            command_d["caltyp2"] = "RKS" if command_d["m"] == 1 else "UKS"
            self.cmd = mydict(command_d)

    def orcasubprocess(self, cmd_string, writeenergy=True, removegbw=True):
        with open(f'{self.wd}/orca.out', 'w') as stdout_file, open(f'{self.wd}/orca_err.out', 'w') as stderr_file:
            p = subprocess.Popen(cmd_string, stdout=stdout_file, stderr=stderr_file, shell=True)
            p.wait()
            rc = p.returncode
        if rc == 0 and writeenergy:
            with open(f'{self.wd}/orca.out', 'r') as f:
                lines = f.readlines()
                for line in lines:
                    if 'FINAL SINGLE' in line:
                        with open(f'{self.wd}/ENERGY', 'w') as f2:
                            f2.write(str(re.findall(r'\-?\d+\.?\d*', line)[0]))
        extensions_to_remove = ['.densities', '.opt']
        keep_list = ['.xyz', '.inp', '.in', '.out', '.input', '.molden', '.gbw']
        if removegbw:
            keep_list.remove('.gbw')
        for filename in os.listdir(f'{self.wd}'):
            if filename == "ENERGY":
                continue
            if not any(filename.endswith(ext) for ext in keep_list):
                os.remove(filename)

        return rc

    def generate_cmd_string(self):
        command_string = "! "
        command_string += f"{self.cmd.get_value('functional', '')} "
        if self.cmd.get_value('basis', ''):
            command_string += f"{self.cmd.get_value('basis', '')} "
        command_string += f"{self.cmd.get_value('caltyp', '')} "
        command_string += f"{self.cmd.get_value('caltyp2', '')} "
        if self.cmd.get_value('solmode', ''):
            command_string += f"CPCM({self.cmd.get_value('solmode', '')}) \n"
        else:
            command_string += f"\n"

        ncpu = self.cmd.get_value('ncpu', 1)
        if isinstance(ncpu, int) and ncpu > 1:
            if self.cmd.get_value('ncpu', 1):
                command_string += f"%PAL NPROCS {int(ncpu)} END \n"
        c = self.cmd.get_value('charge')
        m = self.cmd.get_value('m')
        fn = self.cmd.get_value('fn')
        if all(v is not None for v in [fn, c, m]):
            command_string += f"* xyzfile {c} {m} {fn}.xyz "
        else:
            raise Exception(f"{fn}.xyz in {self.wd} has wrong c,m,fn parameter with {[fn, c, m]}")
        return command_string

    def call(self, removegbw=True):
        command_string = self.generate_cmd_string()
        with open(f'{self.wd}/orca.inp', 'w') as f:
            f.write(command_string)

        cmd = f"{os.environ['ORCA_BIN']} orca.inp"
        rc = self.orcasubprocess(cmd, removegbw=removegbw)
        return rc

    def get_geomres(self):
        if exists(join(self.wd, 'orca.xyz')):
            coord = pd.read_table(join(self.wd, 'orca.xyz'), header=None, skiprows=2, sep='\s+')
        else:
            coord = pd.read_table(join(self.wd, 'coord.xyz'), header=None, skiprows=2, sep='\s+')
        return coord

    def get_energy(self, out="orca.out"):
        Esp = None
        if exists(join(self.wd, "ENERGY")):
            with open(join(self.wd, "ENERGY"), 'r') as file:
                Esp = float(file.readlines()[0])
        elif exists(join(self.wd, out)):
            with open(join(self.wd, out), 'r') as file:
                lines = file.readlines()
            for ln in lines:
                if "FINAL SINGLE POINT ENERGY" in ln:
                    Esp = float(ln.split()[-1])
                    break
        return Esp

    def get_m(self, file="orca.inp"):
        m = None
        if exists(join(self.wd, file)):
            with open(join(self.wd, file), 'r') as file:
                lines = file.readlines()
            for ln in lines:
                if "xyz" in ln:
                    parts = ln.split()
                    m = int(parts[3])
        return m


class CallCrest:
    @classmethod
    def default_settings(cls, setting):
        """Dynamically generates default settings based on the `setting` class."""
        return {"method": setting.get_value("xtb_setting.method", "gfn2"),
                "coordfn": "coord",
                "optlevel": setting.get_value("xtb_setting.optlevel", ""),
                "charge": 0,
                "uhf": 0,
                "alpb": setting.get_value("xtb_setting.alpb", "h2o"),
                "etemp": setting.get_value("xtb_setting.struc_etemp", "4000"),
                }

    def __init__(self, wd, command_d=None, custom_settings=None):
        self.wd = wd
        if command_d is not None:
            self.cmd = mydict(command_d)
        else:
            import aRST.script.initialization as storage
            command_d = self.default_settings(storage.setting)
            if custom_settings is not None:
                command_d.update(custom_settings)

            self.cmd = mydict(command_d)

    def crestsubprocess(self, cmd_string):
        with open(f'{self.wd}/crest.out', 'w') as stdout_file, open(f'{self.wd}/crest_err.out', 'w') as stderr_file:
            p = subprocess.Popen(cmd_string, stdout=stdout_file, stderr=stderr_file, shell=True)
            p.wait()
            rc = p.returncode
        return rc

    def generate_cmd_string(self):
        cmd_string = os.environ['CREST_BIN']+" "
        coordfn = self.cmd.get_value("coordfn")
        if coordfn is None:
            raise Exception(f"coord in {self.wd} can't find right coordfn name")
        cmd_string += f"{coordfn}.xyz "

        cmd_string += f"--{self.cmd.get_value('method')} " \
            if self.cmd.get_value('method') else ""
        cmd_string += f"-c {self.cmd.get_value('charge')} " \
            if self.cmd.get_value('charge') else ""
        cmd_string += f"--uhf {self.cmd.get_value('uhf')} " \
            if self.cmd.get_value('uhf') else ""
        cmd_string += f"--etemp {self.cmd.get_value('etemp')} " \
            if self.cmd.get_value('etemp') else ""
        cmd_string += f"--alpb {self.cmd.get_value('alpb')} " \
            if self.cmd.get_value('alpb') else ""
        cmd_string += f"{self.cmd.get_value('extracmd')} " \
            if self.cmd.get_value('extracmd') else ""
        return cmd_string

    def depotonation(self):
        cmd_string = self.generate_cmd_string()
        cmd_string += " --deprotonate "
        rc = self.crestsubprocess(cmd_string)
        return rc
    def potonation(self):
        cmd_string = self.generate_cmd_string()
        cmd_string += " --protonate "
        rc = self.crestsubprocess(cmd_string)
        return rc

    def read_update_coord(self,job):
        outwd = f'{self.wd}/crest.out'
        with open(outwd, "r") as file:
            lines = file.readlines()
        trigger = False
        startline, end = 0, 0
        for line in lines:
            if "structure    ΔE(kcal/mol)   Etot(Eh)" in line:
                startline = lines.index(line)
                trigger = True
            if trigger and "Wall Time Summary" in line:
                end = lines.index(line) - 3
                trigger = False
                break
        deltaE_l = []
        for line in lines[startline+1:end]:
            nstruc,deltaE,Etot = line.split()
            deltaE_l.append(float(deltaE))
        if min(deltaE_l)>0:
            coord = None
        else:
            nstruc = deltaE_l.index(min(deltaE_l))
            ensemblewd = f'{self.wd}/deprotonated.xyz' if job=="depotonation" else f'{self.wd}/protonated.xyz'
            coord_l = []
            with open(ensemblewd, 'r') as f:

                while True:
                    # Read frame header
                    line = f.readline()
                    if not line:  # End of file
                        break
                    num_atoms = int(line.strip())
                    # Read comment line (contains energy)
                    comment_line = f.readline().strip()

                    # Read atomic coordinates
                    data = []
                    for _ in range(num_atoms):
                        parts = f.readline().split()
                        data.append([parts[0], float(parts[1]), float(parts[2]), float(parts[3])])

                    coord = pd.DataFrame(data)
                    coord_l.append(coord)
            coord = coord_l[nstruc]
        return coord




class CallxTB:
    @classmethod
    def default_settings(cls, setting):
        """Dynamically generates default settings based on the `setting` class."""
        return {"method": setting.get_value("xtb_setting.method", "gfn2"),
                "coordfn": "coord",
                "optlevel": setting.get_value("xtb_setting.optlevel", ""),
                "charge": 0,
                "uhf": 0,
                "alpb": setting.get_value("xtb_setting.alpb", "h2o"),
                "etemp": setting.get_value("xtb_setting.struc_etemp", "4000"),
                }

    def __init__(self, wd, command_d=None, custom_settings=None):
        self.wd = wd
        if command_d is not None:
            self.cmd = mydict(command_d)
        else:
            import aRST.script.initialization as storage
            command_d = self.default_settings(storage.setting)
            if custom_settings is not None:
                command_d.update(custom_settings)

            self.cmd = mydict(command_d)

    def xtbsubprocess(self, cmd_string, writeenergy=False):
        with open(f'{self.wd}/xtb.out', 'w') as stdout_file, open(f'{self.wd}/xtb_err.out', 'w') as stderr_file:
            p = subprocess.Popen(cmd_string, stdout=stdout_file, stderr=stderr_file, shell=True)
            p.wait()
            rc = p.returncode
        if rc == 0 and writeenergy:
            with open(f'{self.wd}/xtb.out', 'r') as f:
                lines = f.readlines()
                for line in lines:
                    if 'TOTAL ENERGY' in line:
                        with open(f'{self.wd}/ENERGY', 'w') as f2:
                            f2.write(str(re.findall(r'\-?\d+\.?\d*', line)[0]))
        return rc

    def generate_cmd_string(self, opt=False, xtbmlversion=False,bar1M=False):
        if xtbmlversion:
            cmd_string = os.environ['XTBML_BIN'] + " "
        else:
            cmd_string = os.environ['XTB_BIN'] + " "
        coordfn = self.cmd.get_value("coordfn")
        if coordfn is None:
            raise Exception(f"coord in {self.wd} can't find right coordfn name")
        cmd_string += f"{coordfn}.xyz "
        if opt:
            optlevel = self.cmd.get_value("optlevel")
            cmd_string += f"--opt {optlevel} " if optlevel else "--opt "

        cmd_string += f"--{self.cmd.get_value('method')} " \
            if self.cmd.get_value('method') else ""
        cmd_string += f"-c {self.cmd.get_value('charge')} " \
            if self.cmd.get_value('charge') else ""
        cmd_string += f"--uhf {self.cmd.get_value('uhf')} " \
            if self.cmd.get_value('uhf') else ""
        cmd_string += f"--etemp {self.cmd.get_value('etemp')} " \
            if self.cmd.get_value('etemp') else ""
        cmd_string += f"--alpb {self.cmd.get_value('alpb')} " \
            if self.cmd.get_value('alpb') else ""
        if bar1M and self.cmd.get_value('alpb'):
            cmd_string += "bar1M "
        cmd_string += f"{self.cmd.get_value('extracmd')} " \
            if self.cmd.get_value('extracmd') else ""
        return cmd_string

    def sp(self, writeenergy=True):
        cmd_string = self.generate_cmd_string(opt=False)
        rc = self.xtbsubprocess(cmd_string, writeenergy)
        return rc

    def opt(self, writeenergy=True):
        cmd_string = self.generate_cmd_string(opt=True)
        rc = self.xtbsubprocess(cmd_string, writeenergy)
        # if rc!=0:
        #     rc = self.secondchance_opt()
        return rc

    def afo(self):
        cmd_string = self.generate_cmd_string(opt=False, xtbmlversion=True)
        cmd_string += "--ml_feature "
        p = subprocess.Popen(cmd_string, stdout=open(f'{self.wd}/xtb.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p.wait()
        rc = p.returncode
        return rc

    def trv(self):
        cmd1 = self.generate_cmd_string()
        cmd1 += "--ohess "
        p1 = subprocess.Popen(cmd1, stdout=open(f'{self.wd}/freq.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p1.wait()
        rc = p1.returncode
        if rc == 0:
            cmd2 = f'{os.environ["TRVT_BIN"]} --io xtb --rotthr 50 '
            cmd2 += f'--fxyz xtbopt.xyz' if os.path.exists(
                join(self.wd, "xtbopt.xyz")) else f'--fxyz {self.cmd.get_value("coordfn")}.xyz '
            cmd2 += f" --temp {self.cmd.get_value('etemp')} " \
                if self.cmd.get_value('etemp') else ""
            cmd2 += f' --fvib freq.out > thermo.out'
            p2 = subprocess.Popen(cmd2, shell=True)
            p2.wait()
            rc2 = p2.returncode
            # if rc2 != 0:
            # raise Exception("TRVthermo calculation failed in",self.wd)
        else:
            rc2 = 1
            # raise Exception("TRV ohess calculation failed in ",self.wd)
        return rc2

    def solv(self):
        cmd2 = self.generate_cmd_string(bar1M=True)
        p2 = subprocess.Popen(cmd2, stdout=open(f'{self.wd}/solv.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p2.wait()
        with open(f'solv.out', 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'TOTAL ENERGY' in line:
                    esolv = float(re.findall(r'\-?\d+\.?\d*', line)[0])
        rc2 = p2.returncode

        oldvalue = self.cmd.get_value("alpb")
        self.cmd.set_value("alpb", False)
        cmd1 = self.generate_cmd_string()
        self.cmd.set_value("alpb", oldvalue)

        p1 = subprocess.Popen(cmd1, stdout=open(f'{self.wd}/gas.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p1.wait()
        rc1 = p1.returncode
        with open(f'{self.wd}/gas.out', 'r') as f:
            lines = f.readlines()
            for line in lines:
                if 'TOTAL ENERGY' in line:
                    egas = float(re.findall(r'\-?\d+\.?\d*', line)[0])
        if rc1 == 0 and rc2 == 0:
            with open(f'{self.wd}/GSOLV', 'w') as f2:
                f2.write(str(esolv - egas))

        return rc2

    def scan(self, setting_d):
        # write inp
        with open(f'{self.wd}/scan.inp', 'w') as f:
            f.write(f'$constrain\n'
                    f'force constant={setting_d["constrain"]["force_constant"]}\n')
            if "loose_bond" in setting_d["constrain"].keys():
                at1, at2, dd = setting_d["constrain"]["loose_bond"]
                f.write('distance: %d,%d,%.2f\n' % (int(at1), int(at2), dd * 1.2))

            f.write('$scan\n')
            if setting_d["scan"]["mode"] == "concerted":
                f.write('mode=concerted\n')

            atl = setting_d["scan"]["distance"]
            for item in atl:
                at1, at2, d0, d_end = item
                f.write('distance: %d,%d,%.2f;%.2f,%.2f,%d\n'
                        % (int(at1), int(at2), d0, d0, d_end, setting_d["scan"]["step"]))
            f.write("$end")

        cmd = self.generate_cmd_string(opt=True)
        cmd += "--input scan.inp "

        p = subprocess.Popen(cmd, stdout=open(f'{self.wd}/xtb.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p.wait()
        rc = p.returncode
        if rc != 0 and (os.path.exists(f'{self.wd}/xtbscan.log') and os.path.exists(f'{self.wd}/xtbopt.xyz')):
            rc = 0
        return rc

    def constrained_opt(self, newbond):
        with open(f'{self.wd}/scan.inp', 'w') as f:
            f.write(f'$constrain\n'
                    f'force constant={self.cmd.get_value("k","1")}\n')
            for item in newbond:
                at1, at2 = item["atoms"]
                value = item["value"]
                f.write('distance: %d,%d,%.2f\n'
                        % (at1, at2, value))
            f.write("$end")
        cmd = self.generate_cmd_string(opt=True)
        cmd += "--input scan.inp "
        p = subprocess.Popen(cmd, stdout=open(f'{self.wd}/xtb.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p.wait()
        rc = p.returncode
        if rc != 0 and os.path.exists(f'{self.wd}/xtbopt.xyz'):
            rc = 0
        return rc

    def secondchance_opt(self):
        cmd_string = "xtb "
        coordfn = self.cmd.get_value("coordfn")
        cmd_string += f"{coordfn}.xyz "
        cmd_string += "--opt "

        cmd_string += f"-c {self.cmd.get_value('charge')} " \
            if self.cmd.get_value('charge') else ""
        cmd_string += f"--uhf {self.cmd.get_value('uhf')} " \
            if self.cmd.get_value('uhf') else ""

        rc = self.xtbsubprocess(cmd_string, True)
        return rc

    def short_md(self, constrained_bonds=False):
        # constrained_bonds = push_l in the scanning
        with open(f'{self.wd}/md.inp', 'w') as f:
            if constrained_bonds:
                f.write("$constrain\n")
                for item in constrained_bonds:
                    at1 = int(item[0])
                    at2 = int(item[1])
                    d = item[3]
                    f.write(f"distance: {at1}, {at2}, {d:.2f}\n")

            f.write("$md\n"
                    f"time={self.cmd.get_value('mdtime', 10)}\n"  # simulation 10 ps 
                    f"step={self.cmd.get_value('mdstep', 1)}\n"  # every 1 fs for propagation
                    f"temp={self.cmd.get_value('mdetemp', 1000)}\n"  # use set relaxed etemp
                    f"shake={self.cmd.get_value('mdshake', 1)}\n"  # not constring bonds
                    f"dump={self.cmd.get_value('mddump', 50)}"  # save geom every 50 fs
                    )
            # xtb coord.xyz --input md.inp --md
        cmd_string = f"xtb {self.cmd.get_value('coordfn')}.xyz --input md.inp --md "
        cmd_string += f"-c {self.cmd.get_value('charge')} " \
            if self.cmd.get_value('charge') else ""
        cmd_string += f"--uhf {self.cmd.get_value('uhf')} " \
            if self.cmd.get_value('uhf') else ""
        cmd_string += f"--alpb {self.cmd.get_value('alpb')} " \
            if self.cmd.get_value('alpb') else ""
        p = subprocess.Popen(cmd_string, stdout=open(f'{self.wd}/xtbmd.out', 'w'), stderr=subprocess.PIPE, shell=True)
        p.wait()
        rc = p.returncode
        return rc

    def get_Emin_geom_from_mdtrj(self, nat):
        import numpy as np

        # cut geomblock in md.trj
        coord_l = []
        energies = []
        trjwd = f'{self.wd}/xtb.trj'

        def parse_xtb_trj(trjwd, nat):
            with open(trjwd, 'r') as f:
                while True:
                    # Read frame header
                    line = f.readline()
                    if not line:  # End of file
                        break
                    num_atoms = int(line.strip())
                    assert num_atoms == nat, f"Unexpected number of atoms: {num_atoms} (expected {nat})"

                    # Read comment line (contains energy)
                    comment_line = f.readline().strip()
                    energy = float(comment_line.split()[1])  # Assumes format: "energy: <value> ..."

                    # Read atomic coordinates
                    data = []
                    for _ in range(num_atoms):
                        parts = f.readline().split()
                        data.append([parts[0], float(parts[1]), float(parts[2]), float(parts[3])])

                    coord = pd.DataFrame(data)
                    coord_l.append(coord)
                    energies.append(energy)
            return coord_l, energies

        def filter_outliers(coord_l, energies, z_threshold=2.5):
            from scipy.stats import zscore
            z_scores = np.abs(zscore(energies))
            non_outliers = np.where(z_scores < z_threshold)[0]

            clean_coord_l = [coord_l[i] for i in non_outliers]
            clean_energies = [energies[i] for i in non_outliers]

            return clean_coord_l, clean_energies

        frames, energies = parse_xtb_trj(trjwd, nat)
        clean_coord_l, clean_energies = filter_outliers(coord_l, energies)
        lowest_idx = np.argmin(clean_energies)
        coord_emin = clean_coord_l[lowest_idx]
        emin = clean_energies[lowest_idx]
        return coord_emin, emin

    def get_moldeninput(self):
        cmd = self.generate_cmd_string()
        cmd += "--molden "
        rc = self.xtbsubprocess(cmd)
        return rc

    def get_charge_from_out(self, out="xtb.out"):
        c = None
        if exists(join(self.wd, out)):
            with open(join(self.wd, out), 'r') as file:
                lines = file.readlines()

            for ln in lines:
                if "net charge" in ln:
                    c = int(ln.split()[-2])
                    break
        return c

    def get_Esp_from_out(self, out="xtb.out"):
        # unit in eh
        Esp = None
        if exists(join(self.wd, "ENERGY")):
            with open(join(self.wd, "ENERGY"), 'r') as file:
                Esp = float(file.readlines()[0])

        elif exists(join(self.wd, out)):
            with open(join(self.wd, out), 'r') as file:
                lines = file.readlines()

            for ln in lines:
                if "total energy" in ln:
                    Esp = float(ln.split()[4])
                    break
        return Esp

    def get_Etrv(self, out="GTRV"):
        # unit in eh
        Etrv = None
        if exists(join(self.wd, out)):
            with open(join(self.wd, out), 'r') as file:
                Etrv = float(file.readlines()[0])
        return Etrv

    def get_Esolv(self, out="GSOLV"):
        # unit in eh
        Esolv = None
        if exists(join(self.wd, out)):
            with open(join(self.wd, out), 'r') as file:
                Esolv = float(file.readlines()[0])
        return Esolv

    def get_topo(self, out="xtbtopo.mol"):
        # constrain_l, ele required in mb; at index starts with 1
        constrain_l, ele = None, None
        coord = None
        for file in os.listdir(self.wd):
            if file.endswith("xyz"):
                from aRST.script.toolbox import read_coord
                coord = read_coord(join(self.wd, file))
                break
        if coord is not None:
            if exists(join(self.wd, out)):
                constrain_l, ele = [], []
                from aRST.script.geom import Coord
                # V2000
                with open(join(self.wd, out)) as f:
                    lines = f.readlines()
                ln_parts = lines[3].split()
                nat, nbond = int(ln_parts[0]), int(ln_parts[1])

                for i in range(nat):
                    ele.append(lines[4 + i].split()[3])
                for i in range(nbond):
                    ln_parts = lines[4 + nat + i].split()
                    at1, at2 = int(ln_parts[0]), int(ln_parts[1])
                    bond_length = Coord(coord).get_ats_distance(at1 - 1, at2 - 1)
                    constrain_l.append({"atoms": [at1, at2], "value": bond_length})
            else:
                # single at
                constrain_l, ele = [], [str(coord.iloc[0, 0])]
        return constrain_l, ele


class CallTerachem:
    _instance = None

    def __new__(cls, wd, Setting=None):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, wd, command_d):
        self.wd = wd
        self.cmd = mydict(command_d)

    def wirte_input(self):
        epsilon = self.cmd.get_value("epsilon")
        coordfn = self.cmd.get_value("coordfn")
        c = self.cmd.get_value("c")
        m = self.cmd.get_value("m")
        functional = self.cmd.get_value("functional")
        basis = self.cmd.get_value("basis")
        if m != 1:
            functional = "u" + functional

        with open(join(self.wd, "tc.inp"), 'w') as f:
            f.write(f"coordinates             {coordfn}.xyz\n"
                    f"method         {functional}\n"
                    f"basis                   {basis}\n"
                    "dftgrid                 1\n"
                    "sphericalbasis          yes\n"
                    "guess                   sad\n"
                    "precision               mixed\n"
                    "genscrdata              true\n"
                    "threall                 1e-14\n"
                    "convthre                1e-4\n"
                    "xtol                    1e-8\n"
                    f"charge                  {c}\n"
                    f"spinmult                {m}\n"
                    )
        if epsilon is None:
            with open(join(self.wd, "tc.inp"), 'a+') as f:
                f.write("run                     energy\n")
        else:
            with open(join(self.wd, "tc.inp"), 'a+') as f:
                f.write("pcm			cosmo\n")
                f.write(f"epsilon			{epsilon}\n")
                f.write("run                     energy\n")

    def run(self):
        cmd = f"{os.environ['TC_BIN']} tc.inp"
        # write input
        self.wirte_input()
        # run
        with open(join(self.wd, "tc.out"), 'w') as stdout_file, open('tc_err.out', 'w') as stderr_file:
            p = subprocess.Popen(cmd, stdout=stdout_file, stderr=stderr_file, shell=True)
            p.wait()
            rc = p.returncode
        return rc


class CallsTDA:
    def __init__(self, wd, coordfn):
        self.wd = wd
        self.coordfn = coordfn

    def callml(self, sty):
        cmd = f"{os.environ['STDA_BIN']} -f {self.coordfn} -ml_feature -sty {str(sty)}"
        p = subprocess.Popen(cmd, stdout=open(join(self.wd, f'stda.out'), 'w'), stderr=subprocess.PIPE, shell=True)
        p.wait()
        rc = p.returncode
        return rc


class CallMultiwfn:
    def __init__(self, wd, inputfn):
        self.wd = wd
        self.inputfn = inputfn

    def write_cartesian_molden(self):
        with open(join(self.wd, "cmd.in"), "w") as f:
            f.write("100\n"
                    "2\n"
                    "6\n"
                    "orca.molden\n"
                    "0\n"
                    "-10\n"
                    )
        cmd = f"{os.environ['MWFN_BIN']} {self.inputfn}"
        with open(join(self.wd, "cmd.in"), "r") as cmd_in_file, open(join(self.wd, "multiwfn.out"), 'w') as out_file:
            p = subprocess.Popen(cmd, stdin=cmd_in_file, stdout=out_file, stderr=subprocess.PIPE, shell=True)
            p.wait()
        rc = p.returncode
        return rc


class CallMolBar:
    def __init__(self, coord):
        self.coord = coord

    def get_mb(self):
        from molbar.barcode import get_molbar_from_coordinates
        coord_l = [self.coord.loc[i, 1:3].tolist() for i in self.coord.index]
        ele_l = [self.coord.loc[i, 0] for i in self.coord.index]
        mb = get_molbar_from_coordinates(coord_l, ele_l)
        return mb



from setuptools import setup, find_packages
import os
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()
setup(
    name = "aRST-chem",
    version = "2.0.1",
    author = "Ying Chen",
    author_email = "ying.chen@pc.rwth-aachen.de",
    description = ("an Automated Reaction Search Tool"),
    license = "MIT",
    url = "https://git.rwth-aachen.de/bannwarthlab/aRST",
    packages=find_packages(),
    exclude_package_data={'': ['.gitignore']},
    include_package_data=True,
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.9',
    ],
    entry_points={
        'console_scripts': [
            'aRST = aRST.script.main:main',
        ]
    }
)



#!/usr/bin/python3.9
# import warnings
# warnings.simplefilter("ignore", UserWarning)
# warnings.simplefilter("ignore", RuntimeWarning)

import os
import shutil
import sys
import time
import traceback
from aRST.script.head import *
from argparse import ArgumentParser
from aRST.script.exploration.plot import plot_rxnnetwork
import aRST.script.initialization as init
from aRST.script.initialization import Setting




def main():

    parser = ArgumentParser(
        # name of program
        prog="aRST",
        # descriptionh
        description="an automated Reaction Search Tool",
        # bottom text
        epilog="By Ying"
    )

    # add arguments

    parser.add_argument("in_file", type=str, help="The input file.")
    parser.add_argument('--explore',
                        '-e',
                        action='store_true',
                        default=False,
                        help='Explore reaction pathes based on given reactants.')
    parser.add_argument('--nanoreactor',
                        '-n',
                        action='store_true',
                        default=False,
                        help='Pseudo-Nanoreactor exploration based on given reactants.')
    parser.add_argument('--afo_features',
                        '-afo',
                        action='store_true',
                        default=False,
                        help='Print out reactive information of given structures.')
    parser.add_argument('--align',
                        '-a',
                        action='store_true',
                        default=False,
                        help='Aligning two strucs with given atom index.')
    parser.add_argument("-u", "--unbuffered", action="store_true",
                        help="Run in unbuffered mode (forces real-time output).")

    args = parser.parse_args()
    if args.unbuffered:
        sys.stdout = open(sys.stdout.fileno(), mode="w", buffering=1)
        sys.stderr = open(sys.stderr.fileno(), mode="w", buffering=1)

    print_head()

    # initialize
    setting = Setting(args.in_file)
    init.initialize(setting)

    print_parser(setting.config)
    print_args(args)

    if args.explore:
        from aRST.script.exploration.assign_reaction import ReactionSearch
        from datetime import datetime
        start_time = datetime.now()
        try:
            ReactionSearch()
            from aRST.script.toolbox import write_rxnout,write_strucout
            write_rxnout()
            write_strucout()
        except Exception as e:
            print(f"Error in main(): {e}", flush=True)
            traceback.print_exc()
        except SystemExit as e:
            print(f"SystemExit: {e}", flush=True)
            traceback.print_exc()
        end_time = datetime.now()
        running_time = end_time - start_time
        total_seconds = running_time.total_seconds()
        print()
        print(f"***** Job finished in {int(total_seconds // 3600)} hour "
              f"{int((total_seconds % 3600) // 60)} mins {total_seconds % 60:.2f} sec *****", flush=True)
        # plot_rxnnetwork()


    elif args.afo_features:
        from aRST.script.reactivity.afo_features import show_afo
        show_afo()
    elif args.align:
        from aRST.script.alignment.do_alignment import aligning_two_structures
        aligning_two_structures()
    elif args.nanoreactor:
        from aRST.script.nanoreactor.build_reaction import pseudo_nanoreactor_inter
        pseudo_nanoreactor_inter()




if __name__ == "__main__":

    main()

